/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1994 - 2000 by Ralf Baechle and others.
 * Copyright (C) 1999 Silicon Graphics, Inc.
 */
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/mman.h>
#include <linux/sys.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/personality.h>

#include <asm/bootinfo.h>
#include <asm/cpu.h>
#include <asm/pgtable.h>
#include <asm/system.h>
#include <asm/mipsregs.h>
#include <asm/processor.h>
#include <asm/stackframe.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/elf.h>
#include <asm/isadep.h>

/* The size of an ELF stack for a process. */
/* NOTE: ELF_STACK_SIZE also defined in fs/binfmt_elf.c */
/* HERE, it must be a power of two.  #-1 used as mask. */
#ifdef CONFIG_CONTIGUOUS_USER_STACK
#define ELF_STACK_SIZE  (64*1024)
#else   /* CONFIG_CONTIGUOUS_USER_STACK */
#define ELF_STACK_SIZE  (64*1024)
#endif  /* CONFIG_CONTIGUOUS_USER_STACK */

void cpu_idle(void)
{
	/* endless idle loop with no priority at all */
	current->nice = 20;
	current->counter = -100;
	init_idle();

	while (1) {
		while (!current->need_resched)
			if (cpu_wait)
				(*cpu_wait)();
		schedule();
#ifndef NO_MM
		check_pgt_cache();
#endif	/* NO_MM */
	}
}

struct task_struct *last_task_used_math = NULL;

asmlinkage void ret_from_fork(void);

void exit_thread(void)
{
	/* Forget lazy fpu state */
	if (last_task_used_math == current) {
		set_cp0_status(ST0_CU1);
		__asm__ __volatile__("cfc1\t$0,$31");
		last_task_used_math = NULL;
	}
}

void flush_thread(void)
{
	/* Forget lazy fpu state */
	if (last_task_used_math == current) {
		set_cp0_status(ST0_CU1);
		__asm__ __volatile__("cfc1\t$0,$31");
		last_task_used_math = NULL;
	}
}

int copy_thread(int nr, unsigned long clone_flags, unsigned long usp,
		 unsigned long stack_size,
                 struct task_struct * p, struct pt_regs * regs)
{
	struct pt_regs * childregs;
	long childksp;

	childksp = (unsigned long)p + KERNEL_STACK_SIZE - 32;

	if (last_task_used_math == current)
		if (mips_cpu.options & MIPS_CPU_FPU) {
			set_cp0_status(ST0_CU1);
			save_fp(p);
		}
	/* set up new TSS. */
	childregs = (struct pt_regs *) childksp - 1;
	*childregs = *regs;
	childregs->regs[7] = 0;	/* Clear error flag */
	if(current->personality == PER_LINUX) {
		childregs->regs[2] = 0;	/* Child gets zero as return value */
		regs->regs[2] = p->pid;
	} else {
		/* Under IRIX things are a little different. */
		childregs->regs[2] = 0;
		childregs->regs[3] = 1;
		regs->regs[2] = p->pid;
		regs->regs[3] = 0;
	}
#if defined(CONFIG_STACK_USED) && defined(CONFIG_CHECKSTACK)
	p->min_stack = 0;
	p->max_stack = 0;
#endif
	if (childregs->cp0_status & ST0_CU0) {
		childregs->regs[28] = (unsigned long) p;
		childregs->regs[29] = childksp;
		p->thread.current_ds = KERNEL_DS;
		p->start_stack = 0;		/* already counted specially */
		p->end_stack = 0;
	} else {
		childregs->regs[29] = usp;
		p->thread.current_ds = USER_DS;
		if (stack_size != 0) {
		  p->start_stack = usp;
printk("copy_thread: stack_size (%ld , 0x%lx)!= 0\n", stack_size, stack_size);
		  p->end_stack = usp - stack_size + 1024;	/* gotta guess */
		} else {
		  unsigned long tmp;
		  unsigned long ptraddr;

		  if ((usp & 0x80000000) == 0) {
		    ptraddr = USER_2_KERN(usp);
		  } else {
printk("copy_thread: can't be here?\n");
		    ptraddr = usp;
		  }
		  if ((tmp = ksize((void *)ptraddr)) == 0) {
printk("copy_thread: ksize returned 0?\n");
		    tmp = ELF_STACK_SIZE;
		  }
/* printk("copy_thread: usp=0x%lx  tmp=%ld (0x%lx)\n", usp, tmp, tmp); */
		  p->start_stack = (usp + tmp) & ~(tmp-1);
		  p->end_stack = usp & ~(tmp-1);
		}
	}

	p->thread.reg29 = (unsigned long) childregs;
	p->thread.reg31 = (unsigned long) ret_from_fork;

	/*
	 * New tasks loose permission to use the fpu. This accelerates context
	 * switching for most programs since they don't use the fpu.
	 */
	p->thread.cp0_status = read_32bit_cp0_register(CP0_STATUS) &
                            ~(ST0_CU2|ST0_CU1|KU_MASK);
	childregs->cp0_status &= ~(ST0_CU2|ST0_CU1);

/* printk("copy_thread(%lx,%lx): pid=%d start_stack=%lx  end_stack=%lx\n", usp, stack_size, p->pid, p->start_stack, p->end_stack); */
	return 0;
}

#if 0
/* Fill in the fpu structure for a core dump.. */
int dump_fpu(struct pt_regs *regs, elf_fpregset_t *r)
{
	/* We actually store the FPU info in the task->thread
	 * area.
	 */
	if(regs->cp0_status & ST0_CU1) {
		memcpy(r, &current->thread.fpu, sizeof(current->thread.fpu));
		return 1;
	}
	return 0; /* Task didn't use the fpu at all. */
}
#endif	/* 0 */

#ifdef CONFIG_BINFMT_AOUT
/* Fill in the user structure for a core dump.. */
void dump_thread(struct pt_regs *regs, struct user *dump)
{
	dump->magic = CMAGIC;
	dump->start_code  = current->mm->start_code;
	dump->start_data  = current->mm->start_data;
	dump->start_stack = regs->regs[29] & ~(PAGE_SIZE - 1);
	dump->u_tsize = (current->mm->end_code - dump->start_code) >> PAGE_SHIFT;
	dump->u_dsize = (current->mm->brk + (PAGE_SIZE - 1) - dump->start_data) >> PAGE_SHIFT;
	dump->u_ssize =
		(current->mm->start_stack - dump->start_stack + PAGE_SIZE - 1) >> PAGE_SHIFT;
	memcpy(&dump->regs[0], regs, sizeof(struct pt_regs));
	memcpy(&dump->regs[EF_SIZE/4], &current->thread.fpu, sizeof(current->thread.fpu));
}
#endif	/* CONFIG_BINFMT_AOUT */

#if 0
/*
 * Create a kernel thread
 */
int kernel_thread(int (*fn)(void *), void * arg, unsigned long flags)
{
	long retval;

	__asm__ __volatile__(
		".set noreorder               \n"
		"    move    $6,$sp           \n"
		"    move    $4,%5            \n"
		"    li      $2,%1            \n"
		"    syscall                  \n"
		"    beq     $6,$sp,1f        \n"
		"    subu    $sp,32           \n"	/* delay slot */
		"    jalr    %4               \n"
		"    move    $4,%3            \n"	/* delay slot */
		"    move    $4,$2            \n"
		"    li      $2,%2            \n"
		"    syscall                  \n"
		"1:  addiu   $sp,32           \n"
		"    move    %0,$2            \n"
		".set reorder"
		:"=r" (retval)
		:"i" (__NR_clone), "i" (__NR_exit),
		 "r" (arg), "r" (fn),
		 "r" (flags | CLONE_VM)
		 /*
		  * The called subroutine might have destroyed any of the
		  * at, result, argument or temporary registers ...
		  */
		:"$1", "$2", "$3", "$4", "$5", "$6", "$7", "$8",
		 "$9","$10","$11","$12","$13","$14","$15","$24","$25");

	return retval;
}
#else
/* The following uses no "s" registers, but assumes v1 is saved and restored */
/* across a system call (syscall).  Quite a bit smaller than above method. */
__asm__ (
	".align	2	\n"
	".globl	kernel_thread	\n"
	".ent	kernel_thread	\n"
	".set	noreorder	\n"
"kernel_thread:	\n"
	"move    $3,$4	\n"
	"ori     $4,$6,0x100	\n"
	"move    $6,$sp		\n"
	"li      $2,120		\n"
	"syscall		\n"
	"beq     $6,$sp,1f	\n"
	"addiu   $sp,$sp,-32	\n"		/* delay slot */
	"jalr    $3		\n"
	"move    $4,$5		\n"		/* delay slot */
	"move    $4,$2		\n"
	"li      $2,1		\n"
	"syscall		\n"
"1:	\n"
	"jr      $ra		\n"
	"addiu   $sp,$sp,32	\n"		/* delay slot */
	".end	kernel_thread	\n"
	".size	kernel_thread, .-kernel_thread	\n"
	".set	reorder"
	);

#endif

/*
 * These bracket the sleeping functions..
 */
extern void scheduling_functions_start_here(void);
extern void scheduling_functions_end_here(void);
#define first_sched	((unsigned long) scheduling_functions_start_here)
#define last_sched	((unsigned long) scheduling_functions_end_here)

/* get_wchan - a maintenance nightmare ...  */
unsigned long get_wchan(struct task_struct *p)
{
	unsigned long frame, pc;

	if (!p || p == current || p->state == TASK_RUNNING)
		return 0;

	pc = thread_saved_pc(&p->thread);
	if (pc < first_sched || pc >= last_sched) {
		return pc;
	}

	if (pc >= (unsigned long) sleep_on_timeout)
		goto schedule_timeout_caller;
	if (pc >= (unsigned long) sleep_on)
		goto schedule_caller;
	if (pc >= (unsigned long) interruptible_sleep_on_timeout)
		goto schedule_timeout_caller;
	if (pc >= (unsigned long)interruptible_sleep_on)
		goto schedule_caller;
/* 	goto schedule_timeout_caller; */
	/* Fall through */

schedule_caller:
	pc = ((unsigned long *)p->thread.reg30)[13];
	return pc;

schedule_timeout_caller:
	/*
	 * The schedule_timeout frame
	 */
	frame = ((unsigned long *)p->thread.reg30)[13];

	/*
	 * frame now points to sleep_on_timeout's frame
	 */
	frame = ((unsigned long *)frame)[9];
	pc    = ((unsigned long *)frame)[10];

	if (pc >= first_sched && pc < last_sched) {
		/* schedule_timeout called by interruptible_sleep_on_timeout */
		frame = ((unsigned long *)frame)[9];
		pc    = ((unsigned long *)frame)[10];
	}

	return pc;
}
