/*
  $Header: /home/cvsroot/B49G/linux-brecis-2.4.20/drivers/brecis/lg79063/SLAClg79063Drv.c,v 1.1.1.1 2004/04/29 09:53:53 jkglee Exp $
*/
/****************************************************************************
 *
 *  Driver for Legerity QSLAC Am 79Q063/061 and SLIC Am 79R79 on BRECIS MSP EVM boards
 *
 *  Note: Does not present an interface to user-level programs,
 *  but implenents the interface defined in brecis_vpm_slac.h to
 *  other driver modules.
 *
 *  Note: this file is no longer responsible for setting up
 *  the TDM clock generator. That functionality resides in the
 *  vpm driver now, this file is responsible for programming
 *  the SLAC to communicate correctly on the TDM bus. To that
 *  end, during initialisation the TDM clock rate is passed
 *  from the calling driver (that presumably set up the TDM
 *  highway.)
 *
 *****************************************************************************/

/******************************************************************/
/* Copyright (c) 2001 BRECIS Communications                       */
/*      This software is the property of BRECIS Communications    */
/*      and may not be copied or distributed in any form without  */
/*      a prior licensing arrangement.                            */
/*                                                                */
/* BRECIS COMMUNICATIONS DISCLAIMS ANY LIABILITY OF ANY KIND      */
/* FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS      */
/* SOFTWARE.                                                      */
/*                                                                */
/******************************************************************/
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/telephony.h>
#include <linux/phonedev.h>
#include "mspCommon.h"
#include <asm/brecis/BrecisSysRegs.h>
#include "msp.h"
#include "SLAClg79063Drv.h"
#include "brecis_vpm_slac.h"
#include "slacpri.h"
#include <brecis/NanoDelay.h>

int loopback=0;	/* module parameter to turn on interface loopback (ILB) */
int tsoffset=0; /* module parameter to offset time slots from 0 */
int clock=-1;	/* module parameter to set non-default clock frequency */

MODULE_PARM(loopback, "i");
MODULE_PARM(tsoffset, "i");
MODULE_PARM(clock, "i");

#ifdef    DEBUG
#define DBG_SLAC(a1, a2...)	printk(KERN_WARNING "SLAC: " a1, ##a2)
#else  /* DEBUG */
#define DBG_SLAC(a...)
#endif /* DEBUG */

/*
 * The board specifics (IRQ number, MPI CS, etc) should not be based on the DEVICE type,
 *  but we have no other way to even guess which board we are running on.  So use the
 *  DEVICE type to decide if we are a 2020, then assume a so-called McKinley configuration.
 */
/* MSP4000/MSP5000 EVMs use IRQ 2, 2020 on McKinley uses IRQ 3 */
static inline int QSLAC_IRQ(void) {return ((DEV_IS_MSP2020()) ? 3 : 2); }

static sSLACInfo   *pSLACInfo= NULL;
static int         slac_max_module_num =0;
static slac_companding_law slac_companding_type = SLAC_MU_LAW;
static int debounce_timer_running =0;
static struct timer_list debounce_timer;

/**********************************************************************
 ** Default SLAC coef. table
 ***********************************************************************/
static slac_coeff_type Am79063DefCoeff = {
	/*AISN: */0x0E,
	{/*GX[2]:*/0x2A, 0x40},
	{/*GR[2]:*/0xA4, 0x21},
	{/*Z[15]:*/0x3C, 0xA2, 0x43, 0xB7, 0x27, 0xAB, 0x32, 0x4C, 0xBB, 0x24, 0xA2, 0x63, 0x97, 0xA8, 0xF0},
	{/*B[14]:*/0xBA, 0xC3, 0xA9, 0xBB, 0x9A, 0xAB, 0x2A, 0xE3, 0xA7, 0xC3, 0x77, 0x97, 0xC9, 0x70},
	{/*X[12]:*/0x01, 0x11, 0x01, 0x90, 0x01, 0x90, 0x01, 0x90, 0x01, 0x90, 0x01, 0x90},
	{/*R[14]:*/0xDC, 0x01, 0x01, 0x11, 0x01, 0x90, 0x01, 0x90, 0x01, 0x90, 0x01, 0x90, 0x01, 0x90},
	{/*B2[2]:*/0x2D, 0x01},
	/*AX:   */0x00,
	/*AR:   */0x00
};

static U8 SlicCtrlByte[SLAC_TYPE_MAX][SLIC_STATE_MAX] =	{
	{
		SLIC_OPEN_CIRCUIT_MODE | SLIC_RING_OFF | SLIC_E1_LOOP_START | SLIC_B2EN_OFF,
		SLIC_RINGING_MODE      | SLIC_RING_ON  | SLIC_E1_LOOP_START | SLIC_B2EN_OFF,
		SLIC_ACTIVE_MODE       | SLIC_RING_OFF | SLIC_E1_LOOP_START | SLIC_B2EN_ON,
		SLIC_OHT_MODE          | SLIC_RING_OFF | SLIC_E1_LOOP_START | SLIC_B2EN_OFF,
		SLIC_TIP_OPEN_MODE     | SLIC_RING_OFF | SLIC_E1_LOOP_START | SLIC_B2EN_OFF,
		SLIC_STANDBY_MODE      | SLIC_RING_OFF | SLIC_E1_LOOP_START | SLIC_B2EN_ON,
		SLIC_APR_MODE          | SLIC_RING_OFF | SLIC_E1_LOOP_START | SLIC_B2EN_ON,
		SLIC_OHT_PR_MODE       | SLIC_RING_OFF | SLIC_E1_LOOP_START | SLIC_B2EN_ON,
	},
	{
		SLIC_OPEN_CIRCUIT_MODE | SLIC_RING_OFF | SLIC_E1_GROUND_START | SLIC_B2EN_OFF,
		SLIC_RINGING_MODE      | SLIC_RING_ON  | SLIC_E1_GROUND_START | SLIC_B2EN_OFF,
		SLIC_ACTIVE_MODE       | SLIC_RING_OFF | SLIC_E1_GROUND_START | SLIC_B2EN_ON,
		SLIC_OHT_MODE          | SLIC_RING_OFF | SLIC_E1_GROUND_START | SLIC_B2EN_OFF,
		SLIC_TIP_OPEN_MODE     | SLIC_RING_OFF | SLIC_E1_GROUND_START | SLIC_B2EN_OFF,
		SLIC_STANDBY_MODE      | SLIC_RING_OFF | SLIC_E1_GROUND_START | SLIC_B2EN_ON,
		SLIC_APR_MODE          | SLIC_RING_OFF | SLIC_E1_GROUND_START | SLIC_B2EN_ON,
		SLIC_OHT_PR_MODE       | SLIC_RING_OFF | SLIC_E1_GROUND_START | SLIC_B2EN_ON,
	},
};

/*serialize access to the SPI MPI bus*/
static spinlock_t spi_mpi_lock = SPIN_LOCK_UNLOCKED;

static void slac_reset_peri_register(void);
static void slac_cfg_ringer_clk( U32 clkFreq);

static void slac_fxs_init(U8 ChipNo,U8 *pTxTSMap,U8 *pRxTSMap);
static void slac_isr(int irq,void* dev_id,struct pt_regs *regs );
static void slac_timeout(unsigned long ptr);
static void slac_add_timer(void);

static U8 QSLACRdCMD_1(U8 ChipNo, U8 Command);

/* 2020 on McKinley uses GPIO 29 as the MPI CS */
#define QSLAC_CS_BIT		(1 << 5)	 /* GPIO 29 data bit */
#define QSLAC_CS_GPIO_DATA_REG  (GPIO_DATA4_REG) /* GPIO 29 data reg */
#define QSLAC_CS_GPIO_CFG_REG   (GPIO_CFG4_REG)  /* GPIO 29 cfg reg */
#define QSLAC_CS_CFG_MASK	(~(0xf << 20))   /* GPIO 29 cfg mask */
#define QSLAC_CS_CFG_VAL	((0x8 << 20))    /* GPIO 29 cfg=output */
#define QSLAC_CS_GPIO_OD_REG    (GPIO_OD_CFG_REG)/* GPIO 29 OD cfg reg */
#define QSLAC_CS_GPIO_OD_BIT    (1 << 29)        /* GPIO 29 OD value 0=active */
#define MPI_CS_ALL_DEASSERTED (SEL_MPI | SPI_MPI_CS0 | SPI_MPI_CS1 | \
	SPI_MPI_CS2 | SPI_MPI_CS3 | SPI_MPI_CS4 | SPI_MPI_CS5 |      \
	SPI_MPI_CS6 | SPI_MPI_CS7 | SPI_MPI_CS8 | SPI_MPI_CS9)

static inline void
slac_assert_chipselect( unsigned int chip)
{
	if (DEV_IS_MSP2020())
	{
		*SPI_MPI_CS_REG &= (~SEL_SPI);			/* make sure we are in MPI mode */
		*QSLAC_CS_GPIO_DATA_REG &= ~QSLAC_CS_BIT;	/* chip select is negative true */
	}
	else
	{
		/*
		 * *assume* individual chip selects are in order, starting
		 *  with bit 0, and negative true (xor to turn bit OFF)
		 */
		*SPI_MPI_CS_REG = MPI_CS_ALL_DEASSERTED ^ (1 << chip);
	}
}

static inline void
slac_deassert_chipselect(void)
{
	if (DEV_IS_MSP2020())
	{
		*QSLAC_CS_GPIO_DATA_REG |= QSLAC_CS_BIT;	/* chip select is negative true */
	}
	else
	{
		*SPI_MPI_CS_REG = MPI_CS_ALL_DEASSERTED;
	}
}

int
slac_open(unsigned int chipNo, unsigned int chanNo)
{
	if ((chipNo != 0) || (chanNo >= pSLACInfo[chipNo].ChansInModule)) {
		return(-ENODEV) ;
	}

	slac_set_state( chipNo,  chanNo, SLIC_ACTIVE);

	pSLACInfo[chipNo].SLICEnable[chanNo] = TRUE;
	pSLACInfo[chipNo].SLICCurrentStatus[chanNo] = ((pSLACInfo[chipNo].SLACLiveStatus) >> 2*chanNo) & 0x01;

	return(0) ;
}

int
slac_close(unsigned int chipNo, unsigned int chanNo)
{
	if ((chipNo != 0) || (chanNo >= pSLACInfo[chipNo].ChansInModule)) {
		return(-ENODEV) ;
	}

	slac_set_state( chipNo,  chanNo, SLIC_STANDBY);
	
	pSLACInfo[chipNo].SLICEnable[chanNo] = FALSE;
	pSLACInfo[chipNo].SLICCurrentStatus[chanNo] = 0x01;

	return(0) ;
}

int
slac_release(unsigned int vpm)
{
	int j ;

	if (vpm != 0) {
		DBG_SLAC("slac_release(), Invalid Slac (%d) !\n",
								vpm) ;
		return (-ENODEV);
	}

	pSLACInfo[vpm].hook_callback = NULL ;
	pSLACInfo[vpm].hook_data = 0 ;


	/*
	 * Flip all channels into open circuit for minimum
	 * power consumption.
	 */
	for (j=0;j<pSLACInfo[vpm].ChansInModule;j++) {
		slac_set_state(vpm, j,SLIC_OPEN_CIRCUIT);
	}
	return(0) ;
}

static void
slac_add_timer(void)
{
	debounce_timer.expires = jiffies + (SLAC_DEBOUNCE_TIMER_INTERVAL * HZ/1000 );
	add_timer(&debounce_timer);
	debounce_timer_running=~0;
}
/**********************************************************************
 ** slac_isr
 **
 ** This is the ISR for QSLAC HW interrupts.
 **
 ***********************************************************************/

static void
slac_isr(int irq,void* dev_id,struct pt_regs *regs )
{
	U8  slacstatus,j;
	U8 chipNo = (U8) (U32) dev_id;
	
	/*Read the data register to clear the interrupt */
	slacstatus =  QSLACRdCMD_1(chipNo, RD_17_RT_DATA_REG_CL) & SLIC_DETECT_MASK ;
	DBG_SLAC("Got QSLAC interrupt on ChipNo: %d, Data Reg 0x%X \n",chipNo,slacstatus);
	if (pSLACInfo[chipNo].SLACLiveStatus != slacstatus) {
		/* update the current status */
		pSLACInfo[chipNo].SLACLiveStatus = slacstatus;
		for (j=0; j<pSLACInfo[chipNo].ChansInModule; j++) {
			if (pSLACInfo[chipNo].SLICCurrentStatus[j] != (slacstatus & 0x01)) {
				/* The status changed: remember current (old) state and set off the debounce*/
				if (pSLACInfo[chipNo].SLICTimerCounter[j] == 0) {
					pSLACInfo[chipNo].SLICCheckStatus[j] = pSLACInfo[chipNo].SLICCurrentStatus[j];
					if (pSLACInfo[chipNo].SLICState[j] == SLIC_RINGING) {
						pSLACInfo[chipNo].SLICTimerCounter[j] = pSLACInfo[chipNo].RingingDBCount[j];
					} else {
						pSLACInfo[chipNo].SLICTimerCounter[j] = pSLACInfo[chipNo].NoRingDBCount[j];
					}
				}
				/* update the current status for this channel */
				pSLACInfo[chipNo].SLICCurrentStatus[j] = slacstatus & 0x01;
			}
			slacstatus = slacstatus >> 2;
		}
	}
} /* slac_isr */

static void
slac_reset_peri_register(void)
{
	int count=1000;
	/*Take the peripheral block out of reset*/
	*RST_CLR_REG = PER_RST;
	while(count--) {
		if (1 == count){printk(KERN_ALERT
			"SLAC: Peripheral block not out of reset after 1 S\n");}
		if (!(*SYS_RST_REG & PER_RST));break;
		NanoDelay(1000000);	/* delay 1 ms = 1000000 ns */
	}

	/*Configure the SPI core control register */
	/*
	*  (1 << 7) Interrupt enabled (hardwired 1)
	 * (00 << 5) clk divide = 8
	 * (01 << 5) clk divide = 16
	 * (10 << 5) clk divide = 32
	 * (11 << 5) clk divide = 64
	 *  (1 << 4) Neg edge flop in sin path (see manual)
	 *  (0 << 4) Neg edge flop in sout path
	 *  (1 << 3) SCK (clock) idles high
	 *  (0 << 3) SCK (clock) idles low
	 *  (1 << 2) Wire-OR mode enabled
	 *  (0 << 2) Wire-OR mode disabled
	 *  (1 << 1) SPI master mode (hardwired 1)
	 * Bit 0 is reserved
	 */
	/* div 8, neg edge flop in sin, clk idles high, no wire-OR */
	*SPI_MPI_CORE_CNTL_REG |= 0x18;
}


static void
QSLACWrCMD(int ChipNo, U8 Command, U8 *pData, U8 Len)
{
	U8  index;
	unsigned long intFlags;

	if (ChipNo >=SLAC_MAX_NUM_OF_CHIP )
		return;

	spin_lock_irqsave(&spi_mpi_lock,intFlags);

	slac_deassert_chipselect();

	*SPI_MPI_CNTL_REG = SPI_MPI_FLUSH_Q;

	if (((*PER_STAT_BIT_REG) & SPI_MPI_FIFO_EMPTY) == SPI_MPI_FIFO_EMPTY) {
		*SPI_MPI_TX_SIZE_REG = 1; /* Transmit MPI QSLAC command only byte-by-byte */
		/* send the READ command */
		*SPI_MPI_DATA_FIFO_REG_U8 = Command ;
		
		slac_assert_chipselect(ChipNo);

		*SPI_MPI_CNTL_REG = SPI_MPI_TX_START ;
		while (((*PER_STAT_BIT_REG) & SPI_MPI_TX_BUSY) == SPI_MPI_TX_BUSY){/*wait*/}
		
		slac_deassert_chipselect();

		/* queue the FIFO buffer */
		for(index = 0; index < Len; index++) { /* size excluded the command */
            /*
			** Poll the receive busy signal but data is written back in FIFO reg
            */
            		if (((*PER_STAT_BIT_REG) & SPI_MPI_FIFO_FULL) != SPI_MPI_FIFO_FULL) {
				*SPI_MPI_DATA_FIFO_REG_U8 = pData[index];
				slac_assert_chipselect(ChipNo);
			
				*SPI_MPI_CNTL_REG = SPI_MPI_TX_START ;
				while (((*PER_STAT_BIT_REG) & SPI_MPI_TX_BUSY) == SPI_MPI_TX_BUSY){/*wait*/}
				
				slac_deassert_chipselect();
			} else {
				DBG_SLAC("SLACWrCMD(), FIFO full\n");
				goto wr_release_lock_return;
			}
		}
	} else {
		DBG_SLAC("SLACWrCMD(), FIFO not empty\n");
		goto wr_release_lock_return;
	}

	*SPI_MPI_CNTL_REG = SPI_MPI_FLUSH_Q;  /* flush Q for next operation */
wr_release_lock_return:
	spin_unlock_irqrestore(&spi_mpi_lock,intFlags);
	return;

} /* QSLACWrCMD */


/**********************************************************************
 ** QSLACRdCMD
 ** NOTE: delay between DeAssertCS and AssertCS must be > 2.5u sec.
 ***********************************************************************/
static void
QSLACRdCMD(U8 ChipNo, U8 Command, U8 *pData, U8 Len)
{
	U8  data=0;
	U8  index;
	unsigned long intFlags;

	if (ChipNo >= 8) {
		DBG_SLAC("QSLACRdCMD invalid ChipNo  %d,\n",ChipNo);
		return;
	}

	spin_lock_irqsave(&spi_mpi_lock,intFlags);

	slac_deassert_chipselect();
	
	*SPI_MPI_TX_SIZE_REG = 1; /* Transmit MPI QSLAC Read command */
	*SPI_MPI_RX_SIZE_REG = 1;/*size - 1 ; Recvd size from QSLAC exclude the command */
	*SPI_MPI_DATA_FIFO_REG_U8 = Command ;
	
	slac_assert_chipselect(ChipNo);
	
	*SPI_MPI_CNTL_REG = SPI_MPI_TX_START ;
	while (((*PER_STAT_BIT_REG) & SPI_MPI_TX_BUSY) == SPI_MPI_TX_BUSY){/*wait*/}
	
	slac_deassert_chipselect();
	
	/* dequeue the FIFO buffer */
	for(index = 0; index < Len; index++) {
		/*
		** Poll the receive busy signal but data is written back in FIFO reg
		*/
		if (((*PER_STAT_BIT_REG) & SPI_MPI_FIFO_FULL) != SPI_MPI_FIFO_FULL) {
			slac_assert_chipselect(ChipNo);
			
			*SPI_MPI_CNTL_REG = SPI_MPI_RX_START ;
			while (((*PER_STAT_BIT_REG) & SPI_MPI_RX_BUSY) == SPI_MPI_RX_BUSY){/*wait*/}
			slac_deassert_chipselect();
			
			data = *SPI_MPI_DATA_FIFO_REG_U8 ;
			pData[index] = data;					
		} else {
			DBG_SLAC("SLACRdCMD(), FIFO full\n");
			goto rd_release_lock_return;
		}
	}

	*SPI_MPI_CNTL_REG = SPI_MPI_FLUSH_Q;  /* flush Q for next operation */
rd_release_lock_return:
	spin_unlock_irqrestore(&spi_mpi_lock,intFlags);
	return;
} /* QSLACRdCMD */


/**********************************************************************
 ** QSLACWrCMD_0
 **
 **
 **
 ***********************************************************************/
static void
QSLACWrCMD_0(U8 ChipNo, U8 Command)
{
	U8     data;

	QSLACWrCMD(ChipNo, Command, &data, 0);
} /* QSLACWrCMD_0 */

/**********************************************************************
 ** QSLACWrCMD_1
 **
 **
 **
 ***********************************************************************/
static void
QSLACWrCMD_1(U8 ChipNo, U8 Command, U8 Data)
{
	U8     data;

	data = Data;
	QSLACWrCMD(ChipNo, Command, &data, 1);
} /* QSLACWrCMD_1 */

/**********************************************************************
 ** QSLACRdCMD_1
 ***********************************************************************/
static U8
QSLACRdCMD_1(U8 ChipNo, U8 Command)
{
	U8     data;

	QSLACRdCMD(ChipNo, Command, &data, 1);
	return(data);
} /* QSLACRdCMD_1 */


/**********************************************************************
 ** SelectChannel
 **
 **
 **
 ***********************************************************************/
static void
SelectChannel(U8 ChipNo, U8 Channel)
{
	/*
	 * robbed bit signaling disabled
	 * Vout = Vref
	 * low power mode OFF
	 */
	QSLACWrCMD_1(ChipNo, WR_14_ENABLE_CH_REG, 1 << Channel);
} /* SelectChannel */

/**********************************************************************
 ** SelectAllChannels
 **
 **
 **
 ***********************************************************************/
static void
SelectAllChannels(U8 ChipNo)
{
	/*
	 * robbed bit signaling disabled
	 * Vout = Vref
	 * low power mode OFF
	 */
	QSLACWrCMD_1(ChipNo, WR_14_ENABLE_CH_REG, ENABLE_ALL_CHANNELS);
} /* SelectAllChannels */

/**********************************************************************
 ** slac_pre_init - Initializes SLAC related data structures
 ***********************************************************************/
static int
slac_pre_init(int NumOfSlac)
{
	int i, j;
	
	if ((NumOfSlac == 0) || (NumOfSlac >= SLAC_MAX_NUM_OF_CHIP)) {
		DBG_SLAC("slac_pre_init(), Invalid NumOfSlac (%d ) !\n", NumOfSlac);
		return (-ENODEV);
	}

	pSLACInfo = (sSLACInfo *)kmalloc(NumOfSlac * sizeof(sSLACInfo),GFP_KERNEL);
	if (NULL == pSLACInfo) {
		DBG_SLAC("slac_pre_init(), Couldn't allocate memory ref!\n" );
		return (-ENOMEM);
	}

	memset(pSLACInfo, 0, NumOfSlac * sizeof(sSLACInfo));

	slac_max_module_num = NumOfSlac;

	for (i=0; i<slac_max_module_num; i++) {
		pSLACInfo[i].ChansInModule = CHANNELS_PER_QSLAC; /*4 channels */
		pSLACInfo[i].SLACLiveStatus = 0xFF;    /*init onhook state !*/
		pSLACInfo[i].OpStatus = FALSE;
		for (j=0; j<pSLACInfo[i].ChansInModule; j++) {
			pSLACInfo[i].SLICEnable[j] = FALSE;
			pSLACInfo[i].SLICCurrentStatus[j] = 0x01;
			/* default configuration info */
			pSLACInfo[i].CircuitType[j] = SLAC_LOOP_START;
			pSLACInfo[i].SlacChannel[j].source = 0;
			pSLACInfo[i].SlacChannel[j].port = j;
			pSLACInfo[i].TimeSlotMap[j] = (j+tsoffset) & 0x7f;
			pSLACInfo[i].NoRingDBCount[j]    = SLAC_DEFAULT_NO_RING_DEBOUNCE_COUNT;
			pSLACInfo[i].RingingDBCount[j]   = SLAC_DEFAULT_RINGING_DEBOUNCE_COUNT;
			pSLACInfo[i].B2TransitDBCount[j] = SLAC_DEFAULT_B2_TRANSIT_DEBOUNCE_COUNT;
			pSLACInfo[i].SLICEnable[j]        = TRUE;
			pSLACInfo[i].SLICCurrentStatus[j] = 0x01;
		}
		
		QSLACWrCMD_0(i, WR_3_HW_RES_REG); /* put HW in known state */
		NanoDelay(1000000); /* wait 1 ms */

		/*
		 * Set all ports to open circuit for minimum power
		 * consumption.
		 */
		for (j=0;j<pSLACInfo[i].ChansInModule;j++) {
			slac_set_state(i, j,SLIC_OPEN_CIRCUIT);
		}
	}
	/*Initialize the debouce timer */
	init_timer(&debounce_timer);
	debounce_timer.function = slac_timeout;
	debounce_timer_running=0;
	slac_add_timer();
	return (0);
}

/*
 * slac_hw_init() -	Initialise the hardware. Called at module
 * 			install time.
 *
 * 			TODO: make this fail with -ENODEV of there
 * 			is no suitable hardware present.
 */
static int
slac_hw_init(void)
{
	int rc ;

	/*Take peripheral block out of reset */
	slac_reset_peri_register();
	if (DEV_IS_MSP2020())
	{
		/* one time setup of GPIO as MPI CS */
		*QSLAC_CS_GPIO_CFG_REG &= QSLAC_CS_CFG_MASK;
		*QSLAC_CS_GPIO_CFG_REG |= QSLAC_CS_CFG_VAL;
		*QSLAC_CS_GPIO_OD_REG  &= ~QSLAC_CS_GPIO_OD_BIT;
		slac_deassert_chipselect();	/* make sure we are in deselected state */
		/*
		 * The GPIO pins must be specifically configured for MPI
		 */
		/* turn off all the associated config bits GPIO 10 and 11 */
		*GPIO_CFG2_REG &= (~((1 << 15) | (1 << 12) | (1 << 11) | (1 << 8)));
		/* turn on the required config bits
		 * (1 << 15) MPI clock (out)
		 * (1 << 12) select MPI clock (as opposed to GPIO)
		 * (1 << 11) MPI data (purportedly a don't care, but make an input)
		 * (1 <<  8) select MPI data (as opposed to GPIO)
		*/
		*GPIO_CFG2_REG |= ((1 << 15) | (1 << 12) | (0 << 11) | (1 << 8));
	}
	
	/* Configure PIT to generate 20 Hz ringer freq on GPIO */
	slac_cfg_ringer_clk(SLIC_RINGER_FREQ);	

	/*Initialize the SLAC data structures*/
	if ((rc = slac_pre_init(1)) != 0) {
		return (rc);
	}

	/* Install handler for SLAC interrupt pass the ChipNo as device id */
	if ((rc =  request_irq(QSLAC_IRQ(), slac_isr, 0, "QSLAC", (void *)0)) != 0) {
		DBG_SLAC("slac_hw_init: Couldn't get IRQ %d, Error: %d\n", QSLAC_IRQ(), rc);
	} else {
		DBG_SLAC("slac_hw_init: Successfully completed SLAC initialization\n");
	}
	return(rc);
}

/**********************************************************************
 ** SlacdrvFxSQSLACFilterSetup
 **********************************************************************/
void
SlacdrvFxSSLACFilterSetup(U8 ChipNo, slac_companding_law compandingLaw, slac_coeff_type *pSlacCoeff)
{

	SelectAllChannels(ChipNo);

	QSLACWrCMD_1(ChipNo, WR_18_AISN_ANA_GAIN, pSlacCoeff->AISN);

	QSLACWrCMD(ChipNo, WR_31_GX_FILTER, &pSlacCoeff->GX[0], 2);

	QSLACWrCMD(ChipNo, WR_33_GR_FILTER, &pSlacCoeff->GR[0], 2);

	QSLACWrCMD(ChipNo, WR_35_Z_FILTER, &pSlacCoeff->Z[0], 15);

	QSLACWrCMD(ChipNo, WR_37_B_FILTER, &pSlacCoeff->B[0], 14);

	QSLACWrCMD(ChipNo, WR_39_X_FILTER, &pSlacCoeff->X[0], 12);

	QSLACWrCMD(ChipNo, WR_41_R_FILTER, &pSlacCoeff->R[0], 14);

	QSLACWrCMD(ChipNo, WR_43_B2_FILTER, &pSlacCoeff->B2[0], 2);

	if (compandingLaw == SLAC_A_LAW) {
		QSLACWrCMD_1(ChipNo, WR_24_OP_REG,
			(COMPRESSED | ALAW_BIT | ENABLE_PROG_FILTERS));
	} else {
		QSLACWrCMD_1(ChipNo, WR_24_OP_REG,
			(COMPRESSED | MULAW_BIT | ENABLE_PROG_FILTERS));
	}

} /* SlacdrvFxSSLACFilterSetup */


/**********************************************************************
 ** slac_fxs_init
 ***********************************************************************/

static inline U8
QSLACConfSetup(void)
{
	/*
	 * turn on the otherwise unused chopper clock bit (6) to provide
	 *  a more unique value for config for detecting readback errors
	 */
	U8 ConfSetup=((QSLAC_CONF_SETUP | (1 << 6)) & 0xf0);
	U8 ClockConstant;

	switch (clock)
	{
		case 1536:
			ClockConstant = 0x0;
			break;
		case 1544:
			ClockConstant = 0x1;
			break;
		case 2048:
			ClockConstant = 0x2;
			break;
		case 3072:
//			ClockConstant = 0x4;
			ClockConstant = 0x0;
			break;
		case 3088:
//			ClockConstant = 0x5;
			ClockConstant = 0x1;
			break;
		case 4096:
//			ClockConstant = 0x6;
			ClockConstant = 0x2;
			break;
		case 6144:
			ClockConstant = 0x8;
			break;
		case 6176:
			ClockConstant = 0x9;
			break;
		case 8192:
			ClockConstant = 0xa;
			break;
		default:
			ClockConstant = 0x1;
			break;
	}
	return (ConfSetup | ClockConstant);
}

void
slac_fxs_init(U8 ChipNo,U8 *pTxTSMap, U8 *pRxTSMap)
{
	volatile U8  ReadBackVal;
	U32 j;

	/*
	 * INTM   (7) = TTL compatible
	 * CHP    (6) = chopper clock 256 KHz
	 * PCMSIG (5) = No signaling on PCM Hwy
	 * CLKSRC (4) = MCLK, no E1
	 * CLK (3..0) = 0001 = 1.544 MHz (implicitly sets # of time slots on PCM
	 */
	QSLACWrCMD_1(ChipNo, WR_12_CONFIG_REG, QSLACConfSetup());

	ReadBackVal = QSLACRdCMD_1(ChipNo, RD_13_CONFIG_REG);
	if (ReadBackVal != QSLACConfSetup()) {
		pSLACInfo[ChipNo].OpStatus = FALSE;
		printk(KERN_ALERT "=====> slac_fxs_init(), RD_CONFIG_REG failed:"
			" ChipNo %d, rdVal=0x%02x != 0x%02x!!\n",
			ChipNo, ReadBackVal, QSLACConfSetup());
		return;
	}
	DBG_SLAC("QSLAC CONFIG REG: 0x%02x (clock %d)\n", ReadBackVal,
								clock);


	QSLACWrCMD_1(ChipNo, WR_22_IO_DIR_REG, SLIC_IO_SETUP);

	/*
	 * TAB = 0 = Transmit on selected PCM Hwy
	 * XE = 1 = Transmit on + edge
	 * RCS = 0 = No RX clock skew
	 * TCS = 0 = No TX clock skew
	 */
	QSLACWrCMD_1(ChipNo, WR_10_CLOCK_EDGE_REG, QSLAC_CLK_EDGE_SETUP);

	for (j=0;j<pSLACInfo[ChipNo].ChansInModule;j++) {
		SelectChannel(ChipNo, j);
		QSLACWrCMD_1(ChipNo, WR_6_TX_TS_SEL_REG, pTxTSMap[j]);
		QSLACWrCMD_1(ChipNo, WR_8_RX_TS_SEL_REG, pRxTSMap[j]);
		DBG_SLAC("QSLAC[%d] TX TS: %d  RX TS: %d\n",
			j, pTxTSMap[j], pRxTSMap[j]);
	}

	if ((clock != 3072) && (clock != 3088) && (clock != 4096)) {
		/* no double PCLK mode, chopper off */
		QSLACWrCMD_1(ChipNo, WR_45_DEBOUNCE_TIME_REG, (0xF<<2)); /*15 ms Debounce */
	} else {
		/* double PCLK mode, chopper off */
		QSLACWrCMD_1(ChipNo, WR_45_DEBOUNCE_TIME_REG, ((0xF<<2)|(1 << 1))); /*15 ms Debounce */
		DBG_SLAC("Double-clock mode enabled %d [0x%02x]\n",
			clock, QSLACRdCMD_1(ChipNo, (WR_45_DEBOUNCE_TIME_REG+1)));
	}

	SelectAllChannels(ChipNo);

	SlacdrvFxSSLACFilterSetup(ChipNo, slac_companding_type, &Am79063DefCoeff );

	QSLACWrCMD_1(ChipNo, WR_28_OP_COND_REG, 0); /* set default */
	if (loopback) {
		QSLACWrCMD_1(ChipNo, WR_28_OP_COND_REG, (1 << 2)); /* I/F Loop Back */
		printk(KERN_INFO "SLAC: ILB ON!\n");
	}

	/* activate chip to turn on PCM Hwy operations (after 2nd frame sync) */
	QSLACWrCMD_0(ChipNo, WR_5_ACTIVATE_REG);

	pSLACInfo[ChipNo].OpStatus = TRUE;

	/* all "A" channels unmasked, all "B" channels masked */
	QSLACWrCMD_1(ChipNo, WR_26_INT_MASK_REG, ~SLIC_INT_MASK);

	for (j=0;j<pSLACInfo[ChipNo].ChansInModule;j++) {
		slac_set_state(ChipNo, j,SLIC_STANDBY);
	}

} /* slac_fxs_init */

/**********************************************************************
 ** SlacInit - Initializes SLAC related data structures
 ***********************************************************************/
int
slac_init(unsigned int vpm, unsigned int nchan, slac_hook_cb *cb, void *data, int clock_khz)
{
	if (NULL == pSLACInfo) {
		DBG_SLAC("slac_init(), Couldn't allocate memory ref!\n" );
		return (-ENOMEM);
	}

	if ((vpm != 0) || (nchan > pSLACInfo[vpm].ChansInModule)) {
		DBG_SLAC("slac_init(), Invalid Slac (%d/%d) !\n",
								vpm, nchan) ;
		return (-ENODEV);
	}

	/*
	 * Unless the user is overriding the clock setting, use
	 * the value passed to us.
	 */
	if (clock == -1) {
		clock = clock_khz ;
	}

	/*
	 * Take note of the callback function we will
	 * call on hook transitions.
	 */
	pSLACInfo[vpm].hook_callback = cb ;
	pSLACInfo[vpm].hook_data = data ;

	slac_fxs_init(vpm, pSLACInfo[vpm].TimeSlotMap,
						pSLACInfo[vpm].TimeSlotMap);

	return (0);
}



/**********************************************************************
 **  slac_set_state
 ***********************************************************************/
int
slac_set_state(unsigned int ChipNo, unsigned int Channel, slac_slic_state NewState)
{
	slac_circuit_type    cirType;
	slac_slic_state      CurrentState;
	U8        CurrentCtrlByte, NewCtrlByte;

	if ((ChipNo != 0) || (Channel >= pSLACInfo[ChipNo].ChansInModule)) {
		return(-ENODEV) ;
	}

	CurrentState = pSLACInfo[ChipNo].SLICState[Channel];

	if (CurrentState == NewState) {
		return(-1);
	}
	DBG_SLAC("SetSLICState: Received State change request 0x%02X  on SLAC %d Channel: %d\n",
			 NewState,ChipNo,Channel);
	cirType = pSLACInfo[ChipNo].CircuitType[Channel];

	SelectChannel(ChipNo, Channel);
	/* write output bits to set SLIC state */
	QSLACWrCMD_1(ChipNo, WR_20_SLIC_IO_REG, SlicCtrlByte[cirType][NewState]);

	pSLACInfo[ChipNo].SLICState[Channel] = NewState;
	CurrentCtrlByte =	SlicCtrlByte[cirType][CurrentState];
	NewCtrlByte = 	SlicCtrlByte[cirType][NewState];

	if ((CurrentCtrlByte & SLIC_B2EN_MASK) != (NewCtrlByte & SLIC_B2EN_MASK)) {
		/* change in B2EN, need debouncing here -- if timercounter already active: overwrite it ! */
		pSLACInfo[ChipNo].SLICCheckStatus[Channel] = pSLACInfo[ChipNo].SLICCurrentStatus[Channel];
		pSLACInfo[ChipNo].SLICTimerCounter[Channel] = pSLACInfo[ChipNo].B2TransitDBCount[Channel];

	}
	return(0) ;
}

static void
slac_timeout(unsigned long ptr)
{
	U8  i, j;
	for (i=0; i<slac_max_module_num; i++) {
		if (pSLACInfo[i].OpStatus == TRUE) {
			for (j=0;j<pSLACInfo[i].ChansInModule;j++) {
				if (pSLACInfo[i].SLICTimerCounter[j] > 0) {
					pSLACInfo[i].SLICTimerCounter[j]--;
					if (pSLACInfo[i].SLICTimerCounter[j] == 0) {
						/* let's check if there is real status change after debouncing */
						if (pSLACInfo[i].SLICCheckStatus[j] != pSLACInfo[i].SLICCurrentStatus[j]) {
							/* yes, status changed -- Current Status is the new status. */
							DBG_SLAC("QSLAC: LOOP %s: %d:%d\n",
								pSLACInfo[i].SLICCurrentStatus[j] ? "Open" : "Closed",
								i,j);
							/*
							 * If there is a callback function registered, call it.
							 */
							if (pSLACInfo[i].hook_callback != NULL) {
								(pSLACInfo[i].hook_callback)(pSLACInfo[i].hook_data, j, pSLACInfo[i].SLICCurrentStatus[j] ? ON_HOOK : OFF_HOOK) ;
							}
							pSLACInfo[i].SLICExceptionFlag[j] = 0xFF;
						}
					}
				}
			}
		}
	}
	slac_add_timer();
}


/*************************************************************************
void SlacCfgRingerClk( U32 clkFreq)	
configures Programmable Interval Timer
to generate a square wave  of (clkFreq) Hz and
connects the output of the timer to GPIO pin.
*************************************************************************/
static void
slac_cfg_ringer_clk( U32 clkFreq)	
{
	if (DEV_IS_MSP2020())
	{
		/* Configure GPIO1 as output for timer counter 1 */
		*GPIO_CFG1_REG &= (~(0x000000F0));	/* don't depend on reset values */
		*GPIO_CFG1_REG |=    0x000000A0;
		*GPIO_OD_CFG_REG &=(~(1 << 1));	/* active  don't depend on reset Values */
		*PER_PIT1_MODE_REG = TIMER_NO_INVERSION | TIMER_MODE_3;
		*PER_PIT1_CNT_REG  = PER_CLOCK_RATE/clkFreq;
		*PER_PIT1_GATE_REG = TIMER_GATE_CONTROL_ON;
	}
	else
	{
		/* Configure GPIO2 as output for timer counter 2 */
		*GPIO_CFG1_REG &= (~(0x00000F00));	/* don't depend on reset values */
		*GPIO_CFG1_REG |=    0x00000B00;
		*GPIO_OD_CFG_REG &=(~(1 << 2));	/* active  don't depend on reset Values */
		*PER_PIT2_MODE_REG = TIMER_NO_INVERSION | TIMER_MODE_3;
		*PER_PIT2_CNT_REG  = PER_CLOCK_RATE/clkFreq;
		*PER_PIT2_GATE_REG = TIMER_GATE_CONTROL_ON;
	}
}

/***************************************************************
SLAC Module Init Function
 - Initializes hardware needed by QSLAC
 - Installs ISR for SLAC interrupt
 - registers QSLAC's as phone devices
***************************************************************/
static int __init
slac_module_init(void)
{
	int rc ;
	
	/*
	 * Perform any configuration-independant init here,
	 * including setting up SPI/MPI interface, configuring
	 * ring oscillator, resetting the slac/slic, programming
	 * the coeffcients and registering irq handler(s.)
	 *
	 * Configuration-specific init should be performed when
	 * slac_init() is called by the vpm driver.
	 */
	if((rc = slac_hw_init()) != 0) {
		DBG_SLAC("slac_hw_init: failed (%d)\n", rc) ;
	} else {
		printk(KERN_INFO "SLAC: "
			"using IRQ %d (" __DATE__ " " __TIME__ ")\n", QSLAC_IRQ());
	}
	return (rc);
}

static void __exit
slac_module_exit( void)
{
	free_irq(QSLAC_IRQ(), NULL );

	if(pSLACInfo) {
		kfree(pSLACInfo);
	}

	del_timer(&debounce_timer);
	pSLACInfo = NULL;

	/*
	 * TODO: generalise
	 */
	QSLACWrCMD_0(0, WR_3_HW_RES_REG); /* put HW in known state */
}


module_init(slac_module_init);
module_exit(slac_module_exit);
MODULE_LICENSE("GPL");

EXPORT_SYMBOL(slac_init) ;
EXPORT_SYMBOL(slac_release) ;
EXPORT_SYMBOL(slac_open) ;
EXPORT_SYMBOL(slac_close) ;
EXPORT_SYMBOL(slac_set_state) ;
