#ifndef	   BRECIS_VPM_SLAC_H_INC_
#define	   BRECIS_VPM_SLAC_H_INC_
/* $Header: /home/cvsroot/B49G/linux-brecis-2.4.20/drivers/brecis/lg79063/brecis_vpm_slac.h,v 1.1.1.1 2004/04/29 09:53:53 jkglee Exp $ */
/* ************************************************************************ *
 *
 * brecis_vpm_slac.h
 *
 *                  Copyright 2003, Brecis Communications
 *
 * Originator: Andy Warner
 *
 *          [Generated by andyw on Thu Aug 14 23:02:39 CDT 2003]
 * ************************************************************************ */

/******************************************************************/
/* Copyright (c) 2003 BRECIS Communications                       */
/*      This software is the property of BRECIS Communications    */
/*      and may not be copied or distributed in any form without  */
/*      a prior licensing arrangement.                            */
/*                                                                */
/* BRECIS COMMUNICATIONS DISCLAIMS ANY LIABILITY OF ANY KIND      */
/* FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS      */
/* SOFTWARE.                                                      */
/*                                                                */
/******************************************************************/

/* The SLAC interface functions.
 *
 * There is a null slac interface, if you want to handle
 * hook control etc some other way.
 *
 * The interface between the SLAC interface and the
 * vpm driver is defined as a set of functions called
 * by the VPM driver, and a single callback function
 * that is registered with the SLAC driver at init
 * time.
 *
 * slac_init() -	Called by the VPM driver to initialise
 *			the slac interface, including h/w init,
 *			registering interrupts etc.
 *			Sets the callback function for hook
 *			state changes.
 *
 *			Arguments are:
 *
 *			vpm - ordinal of the VPM. Currently
 *			only 0 (zero) is supported.
 *
 *			nchan - number of channels specified by
 *			the vpm driver, usually 2, 4, or possibly
 *			8.
 *
 *			callback - address of function to call
 *			on hook state transitions. It is the SLAC
 *			driver's responsibility to debounce.
 *
 * 			data - opaque pointer passed as the first
 * 			argument to to hook status callback function.
 *
 * 			clock_khz - TDM clock rate in KHz. Used to
 * 			assist in programming TDM/filter values.
 *
 * slac_release() -	Called at driver unload time, to release
 *			any resources used, and shutdown the hardware.
 *			Forgets the hook state callback function.
 *
 *			Arguments are:
 *
 *			vpm - ordinal of the VPM. Currently
 *			only 0 (zero) is supported.
 *
 * slac_set_state() -	Called to set the state of a specific
 * 			port.
 *
 *			Arguments are:
 *
 *			vpm - ordinal of the VPM. Currently
 *			only 0 (zero) is supported.
 *
 * 			channel - specifies which port the state
 * 			change applies to.
 *
 * 			state - specifies the desired state, possible
 * 			values are:
 *
 * 			Open Circuit, Ringing, Active, On-Hook Transmission,
 * 			Tip Open, Standby, Active Polarity Reversal,
 * 			On-Hook Transmission Polarity Reversal.
 *
 * slac_hook_cb() -	Callback function passed to slac_init(),
 * 			used to inform vpm driver of hook state
 * 			changes. SLAC driver is responsible for
 * 			debouncing hook transitions. Function
 * 			will be called with the following arguments:
 *
 * 			data - opaque value registered when slac_init()
 * 			is called.
 *
 * 			channel - specifies which port the hook status
 * 			changes on.
 *
 * 			state - debounced hook state (either on-hook
 *			or off-hook.)
 *
 * slac_open() -	Called when the specified port is
 * 			opened.
 *
 * slac_close() -	Called when the specified port is
 * 			closed.
 */

/*
 * This is the definition of the callback function used by
 * the slac interface to inform the driver about hook state
 * changes. When the slac interface is initialised, the vpm
 * driver passes the function
 */
typedef enum {
	ON_HOOK,
	OFF_HOOK,
	FLASH_HOOK
} HOOK_STATE ;

typedef enum {
	SLIC_OPEN_CIRCUIT,
	SLIC_RINGING,
	SLIC_ACTIVE,
	SLIC_OHT,
	SLIC_TIP_OPEN,
	SLIC_STANDBY,
	SLIC_APR,
	SLIC_OHT_PR,
	SLIC_STATE_MAX
} slac_slic_state;

typedef	void (slac_hook_cb)(void *data, unsigned int channel, HOOK_STATE state) ;

extern int slac_init(unsigned int vpm, unsigned int nchan,
			slac_hook_cb *callback, void *data, int clock_khz) ;
extern int slac_release(unsigned int vpm) ;
extern int slac_open(unsigned int vpm, unsigned int channel) ;
extern int slac_close(unsigned int vpm, unsigned int channel) ;
extern int slac_set_state(unsigned int vpm, unsigned int channel,
						slac_slic_state state) ;
#endif	/* BRECIS_VPM_SLAC_H_INC_ */
