/*
 $Header: /home/cvsroot/B49G/linux-brecis-2.4.20/drivers/net/brecismspeth.c,v 1.1.1.1 2004/04/29 09:53:55 jkglee Exp $ 
 */


/******************************************************************
 * Copyright (c) 2002, 2003 BRECIS Communications
 *
 *     This program is free software; you can redistribute it
 *     and/or modify it under the terms of the GNU General
 *     Public License as published by the Free Software
 *     Foundation; either version 2 of the License, or (at your
 *     option) any later version.
 *
 *     This program is distributed in the hope that it will be
 *     useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
 *     02139, USA.
 *
 * BRECIS COMMUNICATIONS DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS
 * SOFTWARE.
 */

/***********************************************************************
 *  brecisethmsp.c : BRECIS EVM ethernet driver for linux
 * 
 * Originally based on MSPEthDrv.c et al, which was a complete re-write
 * by Philip Rakity for the Wind River BSP.
 * 
 * Ideas and code were taken from MaTed mated@lineo.ca when he converted
 * the older code to a Linux driver.

      based on mips BRECIS EVM Board driver for wind river BSP
          Written 2001 by Jey K Surier
        & isa-skeleton.c: A network driver outline for linux.
	    Written 1993-94 by Donald Becker.

		Copyright 2002,2003 BRECIS Communications

	Copyright 1993 United States Government as represented by the
	Director, National Security Agency.

	This software may be used and distributed according to the terms
	of the GNU Public License, incorporated herein by reference.

*************************************************************************/

/************************************************************************
 * modification history
 * --------------------
 *
 * 09/01/01 P Rakity	Complete re-write of original WindRiver driver.
 * 11/15/01 R Sewill	Port of re-written driver.
 ************************************************************************/


#include <asm/brecis/prom.h>
#include <asm/brecis/brecisint.h>

#include <asm/bootinfo.h>

#include "brecismspeth.h"
#include "phyethgeneric.h"

#ifdef CONFIG_NET_FASTROUTE
#include <linux/if_arp.h>
#include <net/ip.h>
#include <net/dst.h>
#endif	/* CONFIG_NET_FASTROUTE */

#ifdef CONFIG_BRECIS_ETH /* skip the entire file if ethernet not enabled */

#define	DRV_NAME	"brecismspeth"
#define	DRV_VERSION	"1.97"
#define	DRV_RELDATE	"Sep 8, 2003"

//#define	tpk	printk
#define	tpk(x...)

static char	*version __devinitdata =
	DRV_NAME ".c:v" DRV_VERSION ", " DRV_RELDATE
	" BRECIS Communications, Corp., www.brecis.com\n";

#define RESERVE2

extern char	*prom_getenv(char *envname);

int __init	brecis_eth_probe(struct net_device *dev);
static int	brecis_eth_probe1(struct net_device *dev, int ioaddr, int unit,
				   unsigned char *ptr);

static void	brecis_eth_tx_timeout(struct net_device *dev);
static void	brecis_eth_InitVars(struct net_device *dev, int ioaddr);
static int	brecis_eth_open(struct net_device *dev);
static int	brecis_eth_send_packet(struct sk_buff *skb, struct net_device *dev);
static void	brecis_eth_interrupt(int irq, void *dev_id, 
					struct pt_regs *regs);
static int	brecis_eth_close(struct net_device *dev);
#if 1
static int	brecis_eth_do_ioctl(struct net_device *dev, struct ifreq *rq,
			int cmd);
#endif
static int	brecis_eth_set_mac_address(struct net_device *dev, void *addr);
static struct net_device_stats	*brecis_eth_get_stats(struct net_device *dev);
static void	set_multicast_list(struct net_device *dev);

static void	MSPEthMemoryFree(MSP_DEVICE *pDrvCtrl);

/* #define DEBUGDISPLAYMACREGISTERS */

static void	MSPEthPhyInit(MSP_DEVICE *pDrvCtrl);
static void	MSPEthPhyInit_complete(MSP_DEVICE *pDrvCtrl);
static void	brecis_eth_complete_open(MSP_DEVICE *priv);
static void	MSPEthSetDuplex(MSP_DEVICE *pDrvCtrl, int duplexMode);
static void	MacSetupRegistersTx(MSP_DEVICE *pDrvCtrl);
static void	MacSetupRegistersRx(MSP_DEVICE *pDrvCtrl, int fdaSize);
static void	RXbh(void *data);
static void	reload_tq(void *data);
#ifdef ETHERNETUSESCRATCHRAM
static u8	*kmallocSRAM(int size);
static void	kfreeSRAM(void *);
#endif
static u8	*kmallocBDS(int size);
static void	kfreeBDS(void *);
static u8	*memalign16(u8 *);

static void	reloadchecks(unsigned long);
static void	reloadcheck_timer(unsigned long);
static	DECLARE_TASKLET(mspeth_reloadcheck_task, reloadchecks, 0);
static struct timer_list	mspeth_reloadcheck_timer;

static void	checkphys(unsigned long);
static void	checkphys_timer(unsigned long);
static	DECLARE_TASKLET(mspeth_checkphys_task, checkphys, 0);
static void	monitor_phy_negotiation(void);

#ifdef	CONFIG_NET_FASTROUTE
static int	msp_accept_fastpath(struct net_device *dev,
			struct dst_entry *dst);
#endif	/* CONFIG_NET_FASTROUTE */

//#define	SKB_RECYCLE	1

#ifdef	SKB_RECYCLE
#define	MSP_RECYCLE_LIST_SIZE	200	/* Max # private buffers */

u32	recycle_hits, recycle_misses;

struct sk_buff_head msp_recycle_list;

static inline void	msp_recycle_skb(struct sk_buff *skb)
{
	if (atomic_read(&skb->users) == 1 && skb->destructor == 0 &&
		skb->truesize >= (MSP_END_BUFSIZE + 15) &&
	/*if (skb->type == 1 && */
		skb_queue_len(&msp_recycle_list) <= MSP_RECYCLE_LIST_SIZE) {
		skb->head = skb->data;
		skb->tail = skb->data;
		skb->len = 0;
		/* LIFO queue */
		tpk("+");
		skb_queue_head(&msp_recycle_list, skb);
	} else {
		tpk("f%d", skb->truesize);
		dev_kfree_skb_any(skb);
	}
}
#endif	/* SKB_RECYCLE */

static struct timer_list	mspeth_checkphys_timer;

/*
 * The name of the card. Is used for messages and in the requests for
 * io regions, irqs and dma channels
 */
static const char	*cardname = "brecismspeth";
static char	brctl_present;		/* True if brctl register present */

MSP_DEVICE	* volatile gEthDrvCtrl[MSP_ETH_MAX_UNITS];
static struct tq_struct	rx_bh[MSP_ETH_MAX_UNITS];
static struct tq_struct	reload_tq_g[MSP_ETH_MAX_UNITS];
static char	unit_init[MSP_ETH_MAX_UNITS];

#ifdef DEBUG_RX_TX_STATS
static unsigned int	fdaExhausedCounts[MSP_ETH_MAX_UNITS];
static unsigned int	blExhausedCounts[MSP_ETH_MAX_UNITS];
#endif


/* #define DEBUGINTCOUNT	(ETH_RX_BD_COUNT+1) */
#ifdef DEBUGINTCOUNT
#define MAX_DEVICES	3
int	rxpackets[MAX_DEVICES][DEBUGINTCOUNT];
int	intcount[MAX_DEVICES];
#endif

/******************************************************
 * Internal brecismspeth data structures
 ******************************************************/
// 2 things to remember - use the KSEG1 address to hardware
//                      - is the ordering of the indices correct
static const u32  	MacBases[MSP_ETH_MAX_UNITS + 1] =
  { MAC0_BASE, MAC1_BASE, MAC2_BASE, 0};
static const u32	MacRSTBases[MSP_ETH_MAX_UNITS + 1] =
{ MAC0_RST, MAC1_RST, MAC2_RST, 0};
static const unsigned char 	MacIRQs[MSP_ETH_MAX_UNITS + 1] = 
  { BRECISINT_MAC0, BRECISINT_MAC1, BRECISINT_MAC2, 0}; 
static PHY_BASE		PhyBases[MSP_ETH_MAX_UNITS + 1] =
  { {SPIN_LOCK_UNLOCKED, (MAC_BASE *) MAC0_BASE}, 
    {SPIN_LOCK_UNLOCKED, (MAC_BASE *) MAC1_BASE}, 
    {SPIN_LOCK_UNLOCKED, (MAC_BASE *) MAC2_BASE}, 
    {SPIN_LOCK_UNLOCKED, 0}};

static u32	phyAddresses[MSP_ETH_MAX_UNITS];
static u8	mspeth_inited;

static short	gethspeedduplex[MSP_ETH_MAX_UNITS];	/* duplex of eth interface */
static unsigned char	gethswitch[MSP_ETH_MAX_UNITS];	/* non-zero if eth switch */

static char	*MacName[]
	= {"ethaddr0", "ethaddr1", "ethaddr2", "unknown"};

#ifdef ETHERNETUSESCRATCHRAM
#define SCRATCH_LAST	0	/* must be zero */
#define SCRATCH_FREE	1
#define SCRATCH_INUSE	2

static unsigned long	scratchramaddr[MAXSCRATCHRAMPARTITIONS + 1] =
{ SCRATCHRAMADDR, SCRATCHRAMADDR + ETHERNETSCRATCHRAMLEN};

static unsigned long	scratchramstate[MAXSCRATCHRAMPARTITIONS + 1] =
{ SCRATCH_FREE, SCRATCH_LAST };
#endif

// #define DEBUGJIFFIEINT
#ifdef DEBUGJIFFIEINT
#define MAXINTSPERJIFFIE	10000
static int debuginterruptcount;
static int debugjiffiecount;
int debugintdev[MAXINTSPERJIFFIE];
int debugintstat[MAXINTSPERJIFFIE];
#endif


/***************************************************************************


 +--------+-----------+---------+---------+
  Byte 3     Byte 2     Byte 1     Byte 0
 +--------+-----------+---------+---------+
            FRAME DESCRIPTOR

+------------------------------------------+
|               FDNext			   | 0x00
+------------------------------------------+
|              FDSystem			   | 0x04
+------------------------------------------+
|               FDStat			   | 0x08	
+------------------+-----------------------+
|   FDCtl          |          FDLength     | 0x0c
+------------------+-----------------------+




            BUFFER DESCRIPTOR
+------------------------------------------+
|    BuffData (pointer to the buffer)      | 0x00
+------------------------------------------+
|              BStatus                     | 0x04
+------------------------------------------+


*****************************************************************************/

static inline void InvalidateFDBD(FDBD *fdbd)
{
#ifdef ETHERNETUSESCRATCHRAM
	if ((unsigned long) fdbd >= SCRATCHRAMADDR &&
	    (unsigned long) fdbd < (SCRATCHRAMADDR + ETHERNETSCRATCHRAMLEN))
	{
		return;
	}
#endif
#if MAX_BD_PER_FRAME == 2
	invalidate_dcache_line((u32) fdbd);
	invalidate_dcache_line(((u32) fdbd) + L1_CACHE_BYTES);
#else
	InvalidateBuffer(fdbd, sizeof(FDBD));
#endif
}

#ifdef	SKB_RECYCLE
static inline struct sk_buff *getskb(MSP_DEVICE *pDrvCtrl)
{
	struct sk_buff *skb2 = 	skb_dequeue(&msp_recycle_list);

	if (skb2 == NULL) {
		skb2 = dev_alloc_skb(MSP_END_BUFSIZE + 15);
		skb2->dev = pDrvCtrl->netdevice;
		skb_reserve(skb2, 2 + MSP_END_PREPEND);
		recycle_misses++;
		tpk("A");
	} else {
		skb_reserve(skb2, 16 + 2 + MSP_END_PREPEND);
		tpk("-");
		recycle_hits++;
	}
	return skb2;
}
#else
static inline struct sk_buff	*getskb(MSP_DEVICE *pDrvCtrl)
{
	struct sk_buff *skb;

	skb = dev_alloc_skb(MSP_END_BUFSIZE + 15);

	if (skb != NULL)
	{
		skb->dev = pDrvCtrl->netdevice;
#ifdef RESERVE2
		/* The symbol RESERVE2 is defined.
		 * The way our hardware works is as follows:
		 * 1) the BlFrmPtr.BuffData pointer must be 32 bit word
		 *    aligned.
		 * 2) We offset the buffer by 2 to get the IP header
		 *    32 bit aligned.
		 * 3) When RESERVE2 is defined, the DMA_CTL register bit
		 *    DMA_CTL_RxAlignSkip2 is defined causes the hardware
		 *    to advance two bytes into the buffer when receiving
		 *    a packet.
		 */
		skb_reserve(skb, 2 + MSP_END_PREPEND);
#endif
	}

	return skb;
}
#endif	/* SKB_RECYCLE */


static inline u8 *getRxBuffData(UCHAR *BuffData)
{
	return ((u8 *) (((unsigned long) BuffData) & ~0x3));
}

static inline int MACOwnsTxFD(FDBD *fdbd)
{
	InvalidateFDBD(fdbd);

	return ((fdbd->FD.FDCtl & MAC_OWNS_FD) || 
		! (fdbd->FD.FDStat & TX_STAT_Comp));
}

static inline int TxFDInUse(FDBD *fdbd)
{
	return fdbd->FD.FDSystem != 0;
}

static inline int MACOwnsRxFD(FDBD *fdbd)
{
	InvalidateFDBD(fdbd);

	return fdbd->FD.FDCtl & MAC_OWNS_FD;
}


static inline void setup_delayed_reload(struct net_device *dev)
{
	MSP_DEVICE	*priv = dev->priv;

	printk(KERN_ERR "setup_delayed_reload: dev<%x>->name<%s>\n",
		   (unsigned int) dev, dev->name);

	netif_carrier_off(dev);
	netif_stop_queue(dev);
	priv->reloadflag = 1;
}


/*********************************************************************
 * Function Name:  MacSetupTxBDs
 *
 * Purpose:	To allocate memory for transmit frame and buffer descriptors
 *		and set up initial values for these.
 *
 * Inputs:	interface number: 0 or 1.
 *
 * Return Value:
 *		MSP_SUCCESS
 *		MSP_MAC_MEM_ALLOC_ERROR
 */

static u32	MacSetupTxBDs(MSP_DEVICE *pDrvCtrl)
{
	int	i;
	FDBD	*tempFdBd;
	u32	txBdCount = pDrvCtrl->txBdCount;
 
	PRINTK(KERN_INFO "MacSetupTxBDs: Setting up TX Buffer Descriptors\n");
	PRINTK(KERN_INFO "MAC# %d, txBdCount: %d\n", 
		pDrvCtrl->unit, (UINT) txBdCount);

	/*
	 * Set up the TXQ. The TXQ will have MAX_BD_PER_FRAME BDs hanging off 
	 * of each FD. There will be txBdCount of FDBD. 
	 * But to make sure 16 byte aligned for each FD, the MAX_BD_PER_FRAME 
	 * should be even number.
	 * Allocate memory for transmit frame and buffer descriptors.
	 */
	if (!(pDrvCtrl->TxFrmPtr_Orig = kmallocBDS(txBdCount * sizeof(FDBD)))) 
	{ 
		printk(KERN_INFO "MacSetupTxBDs: ERROR Can't allocate memory for TxQ!!");
		return MSP_MAC_MEM_ALLOC_ERROR;
	}

	pDrvCtrl->TxFrmPtr = (FDBD *) memalign16(pDrvCtrl->TxFrmPtr_Orig);

	memset(pDrvCtrl->TxFrmPtr, 0, txBdCount * sizeof(FDBD));

	tempFdBd = pDrvCtrl->TxFrmPtr;

	for (i = 0; i < txBdCount; i++) 
	{  
		FDBD	*temp;

		tempFdBd->FD.FDCtl	= BD_PER_TX_FRAME;
		temp			= tempFdBd;
		temp++;
		tempFdBd->FD.FDNext	= &(temp->FD);
		tempFdBd		= temp;
	}	

	/* Set the FDNext in the last FD, to point to the first. */
	tempFdBd--;
	tempFdBd->FD.FDNext		= &(pDrvCtrl->TxFrmPtr->FD); 


	/* Set the TxFrmPtrCur to the base of the TXQ. */
	pDrvCtrl->TxFrmPtrCur		= pDrvCtrl->TxFrmPtr;
	pDrvCtrl->TxFrmPtrFreeBuf	= pDrvCtrl->TxFrmPtr;

	PRINTK(KERN_INFO "\nTxFrmPtr[%d]: 0x%x\n",
			pDrvCtrl->unit, (UINT) pDrvCtrl->TxFrmPtr);
	PRINTK(KERN_INFO "TxFrmPtrCur[%d]: 0x%x\n", 
			pDrvCtrl->unit, (UINT) pDrvCtrl->TxFrmPtrCur);

	return MSP_SUCCESS;
}


/****************************************************************************
 *  Function Name:	MacSetupRxBDs                              
 *                                                                          
 *  Purpose:        Prepare memory for receive buffers and frame descriptors
 *                  and buffer descriptors.                                
 *                                                                         
 *  Inputs:			
 *	rxBdCount	 - Receive Buffer Descriptor Count. In this version 
 *			   of the driver, each BD points to one Rx Buffer, 
 *			   hence rxBdCount is equal to the receive buffer count.
 *	rxBuffCount	 - Receive Buffer Count.
 *
 *  Return Values:
 *	MSP_SUCCESS
 *	MSP_MAC_MEM_ALLOC_ERROR
 */

static u32	MacSetupRxBDs(MSP_DEVICE *pDrvCtrl) 
{
	int		i;
	u32		fdaSize;
	u32		blSize;
	u32		fdaCount;	
	FD		*tempFd;
	struct sk_buff	*skb;

	PRINTK(KERN_INFO "MacSetupRxBDs: Setting up Rx Buffers and Descriptors\n");
	PRINTK(KERN_INFO "MAC# %d, rxBdCount: %d\n", 
		pDrvCtrl->unit, (UINT)pDrvCtrl->rxBdCount);

	/*
	 * Buffer List: (which consisits of single FD and multiple BDs)
	 * There will be rxBdCount counts of BDs, and each BD points to one 
	 * RecBuffSpace of size of MAX_BUFF_SIZE.
	 */
	blSize = pDrvCtrl->rxBdCount * sizeof(BD) + sizeof(FD);
	if (!(pDrvCtrl->BLFrmPtr_Orig = kmallocBDS(blSize)))
	{ 
		printk(KERN_WARNING
			"MacSetupRxBDs: ERROR Can't allocate memory for BL!!\n");
		return MSP_MAC_MEM_ALLOC_ERROR;
	}
	
	pDrvCtrl->BLFrmPtr = (BLFDBD *)memalign16(pDrvCtrl->BLFrmPtr_Orig);

	memset(pDrvCtrl->BLFrmPtr, 0, blSize);

	/*
	 * Allocate memory for (FDA)descriptors for received buffers.
	 * For each frame received, one FD+BD will be written in the FDA if the
	 * rx buffer that is pointed by the BD could fit the entire packet.
	 * The maximum space need in the free area is 1FD + 1BD rounded by to 
	 * a multiple of the FD size.  So, since 2BD's == 1FD, the minimum
	 * size a good receive packet will take is 2 FD's worth of space.
	 *
	 * WARNING ==  It is important that FdaCount NOT exceed the buffer
	 * limit number for rx buffers.  Eg.. we have an FDA Exhausted
	 * interrupt and NOT a BLExhausted Interrupt... 
	 */
	fdaCount = 2 * pDrvCtrl->rxBdCount;
	fdaSize = fdaCount * sizeof(FD);
	if (!(pDrvCtrl->FDAPtr_Orig = kmallocBDS(fdaSize)))
	{ 
		printk(KERN_WARNING
			"MacSetupRxBDs: ERROR Can't allocate memory for FDA!!");
		return MSP_MAC_MEM_ALLOC_ERROR;
	}

	pDrvCtrl->rskbp = 
		kmalloc(sizeof(struct sk_buff *) * pDrvCtrl->rxBdCount,
			GFP_KERNEL);

	if (pDrvCtrl->rskbp == NULL)
	{
		printk(KERN_WARNING
			"MacSetupRxBDs: ERROR Can't allocate memory for rskbp!!");
		return MSP_MAC_MEM_ALLOC_ERROR;
	}

	pDrvCtrl->FDAPtr = (FDBD *) memalign16(pDrvCtrl->FDAPtr_Orig);

	memset(pDrvCtrl->FDAPtr, 0, fdaSize);

	pDrvCtrl->FDAPtrCur = pDrvCtrl->FDAPtr;

	/*
	 * Initialize the FDA Area. Give the ownership of each 16 byte block
	 * to the controller. (The patterns of  0xBBBBBBBB etc is for debugging only!)
	 */
	tempFd = &(pDrvCtrl->FDAPtr->FD);
	for (i = 0; i < fdaCount; i++)
	{
		tempFd->FDSystem = 0xBBBBBBBB;
		tempFd->FDStat   = 0xCCCCCCCC;
		tempFd->FDCtl    = MAC_OWNS_FD;
		tempFd->FDLength = 0xEEEE;
		tempFd->FDNext   = &tempFd[1];
		tempFd++;
	}
	
	/* Last FD points to first */
	tempFd--;
	tempFd->FDNext   = &(pDrvCtrl->FDAPtr->FD);


	/*
	 * Initialize the BL Area. We have only one FD with rxBdCount BDs here,
	 * so point to self as FDNext
	 */
	pDrvCtrl->BLFrmPtr->FD.FDNext	 = &(pDrvCtrl->BLFrmPtr->FD);
	pDrvCtrl->BLFrmPtr->FD.FDSystem	 = 0xAAAAAAAA;	/* debugging  */
	
	/*
	 * Set the number of BDs in the FDLength field of BLFrmPtr 
	 * according to the spec.
	 *
	 * Controller owns the FD, BDCount left blank as the controller
	 * uses FDLength to show the number of BDs available in the Buffer
	 * List queue, but FDLength is set to rxBdCount buffer descriptors 
	 */
	pDrvCtrl->BLFrmPtr->FD.FDCtl	 = MAC_OWNS_FD;
	pDrvCtrl->BLFrmPtr->FD.FDLength  = pDrvCtrl->rxBdCount;	


	/*
	 * Each BD points to a large enough buffer so that for each 
	 * incoming packet
	 * only one BD per packet will be used. So, the number of BD should be 
	 * the same as number of the receive buffers which is rxBuffCount
	 */
	for (i = 0; i < pDrvCtrl->rxBdCount; i++) 
	{
		unsigned long	intlock;

		save_and_cli(intlock);
		
		/* Get a Buffer and attach to the BD.*/
		skb = getskb(pDrvCtrl);

		restore_flags(intlock);

		if (skb == NULL)
		{
			printk(KERN_INFO
				"MacSetupRxBDs: skb NULL for MAC%d, on Cnt=%d\n", 
				pDrvCtrl->unit, i);
			return MSP_MAC_MEM_ALLOC_ERROR;
		}

		pDrvCtrl->rskbp[i] = skb;
			
		/* Place index of sk_buff structure in BDStat */

		pDrvCtrl->BLFrmPtr->BD[i].BuffData = getRxBuffData(skb->data);
		pDrvCtrl->BLFrmPtr->BD[i].BDCtl	= MAC_OWNS_BD;
		pDrvCtrl->BLFrmPtr->BD[i].BDStat = i;
		pDrvCtrl->BLFrmPtr->BD[i].BuffLength = MSP_END_BUFSIZE;
	}

#ifdef DEBUGPRINTK
	{
		unsigned int *ptr = (unsigned int *) pDrvCtrl->BLFrmPtr;

		PRINTK(KERN_INFO "BLFrmPtr[%d] %x, %08x, %08x, %08x, %08x\n", 
			pDrvCtrl->unit, (UINT) ptr,
			ptr[0], ptr[1], ptr[2], ptr[3]);

		ptr = (unsigned int *) &pDrvCtrl->BLFrmPtr->BD[0];

		PRINTK(KERN_INFO "BLFrmPtr[%d]->BD[0] = %08X, %08x, %08x\n", 
			pDrvCtrl->unit, 
			(UINT) ptr, ptr[0], ptr[1]);

		ptr = (unsigned int *) &pDrvCtrl->BLFrmPtr->BD[1];

		PRINTK(KERN_INFO "BLFrmPtr[%d]->BD[1] = %08X, %08x, %08x\n", 
			pDrvCtrl->unit, 
			(UINT) ptr, ptr[0], ptr[1]);

		i--;

		ptr = (unsigned int *) &pDrvCtrl->BLFrmPtr->BD[i];

		PRINTK(KERN_INFO "BLFrmPtr[%d]->BD[%d] = %08X, %08x, %08x\n", 
			pDrvCtrl->unit, i, 
			(UINT) ptr, ptr[0], ptr[1]);
	}
#endif
	MacSetupRegistersRx(pDrvCtrl, fdaSize);
#ifdef DEBUGPRINTK
	{
		unsigned int *ptr = (unsigned int *) pDrvCtrl->FDAPtrCur;

		PRINTK(KERN_INFO "FDAPtrCur[%d] %x, %08x, %08x, %08x %08x\n", 
		       pDrvCtrl->unit, (UINT) ptr,
		       ptr[0], ptr[1], ptr[2], ptr[3]);

		PRINTK(KERN_INFO "     %08x, %08x, %08x, %08x\n",
		       ptr[4], ptr[5], ptr[6], ptr[7]);

		ptr = (unsigned int *) tempFd;

		PRINTK(KERN_INFO "tempFd[%d] %x, %08x, %08x, %08x %08x\n", 
		       (UINT) fdaCount - 1, (UINT) ptr,
		       ptr[0], ptr[1], ptr[2], ptr[3]);

		PRINTK(KERN_INFO "     %08x, %08x, %08x, %08x\n",
		       ptr[4], ptr[5], ptr[6], ptr[7]);
	}
#endif
	return MSP_SUCCESS;
}


/*
 * msp_tx_process_error - Process transmit errors
 *
 * This function should be called infrequently, so separating it
 * gets it out of the hot section.
 */

static void	msp_tx_process_error(MSP_DEVICE *priv, u32 status)
{
	priv->stats.tx_errors++;

	/* Now log specific error */
	if (status & (TX_EX_COLL + TX_LATE_COLL))
		priv->stats.collisions++;

	if (status & (TX_DEFER + TX_PAUSED + TX_EXDEFER + TX_HALTED))
		priv->stats.tx_aborted_errors++;
	
	if (status & TX_EXDEFER)
		priv->stats.tx_carrier_errors++;

	if (status & (TX_PARITY + TX_UNDER))	// Underrun is here??
		priv->stats.tx_fifo_errors++;

	if (status & TX_SQERR)
		priv->stats.tx_heartbeat_errors++;
	
#if 0	// There are circumstances when the transmitter has to be restarted
	/*	if (status & (TX_PARITY + ???))
		// Restart transmitter */
#endif
}


/*************************************************************************
 * MSPEthFreeTxBuffs -	This function looks for any transmit completed
 *			Descriptors and frees up the buffer associated
 *			with them.
 *
 * RETURNS: N/A.
 */

static void __hotnet	MSPEthFreeTxBuffsInternal(MSP_DEVICE *priv)
{
	FDBD		*fdbd;	
	unsigned long	flags;
	struct sk_buff	*skb;
	u32		status;		
	int		wakequeue = 0;

	spin_lock_irqsave(&priv->lock, flags);

	/* 
	 * If the FDStat field is non-zero, then it indicates a transmission. 
	 * Go and find all such FDs and frees up the buffers that have been 
	 * already transmitted.
	 */
	/* Free up the Tx Buffers */
	do
	{
		fdbd = priv->TxFrmPtrFreeBuf;

		priv->TxFrmPtrFreeBuf = (FDBD *)fdbd->FD.FDNext;

		skb = (struct sk_buff *)fdbd->FD.FDSystem;

		/*
		 * Gather statistics
		 */
		status = fdbd->FD.FDStat;

#if 0
		PRINTK(KERN_INFO "Freeing %x skb %x buf %x len %x status %x\n",
		       (unsigned int) fdbd, (unsigned int) skb,
		       (unsigned int) skb->data, skb->len, 
		       (unsigned int) status);
#endif

		if (unlikely(status & TX_STAT_ANYERROR))
			msp_tx_process_error(priv, status);
		else
		{
			priv->stats.tx_bytes += skb->len;	// Total bytes transmitted
			priv->stats.tx_packets++;		// Total packets transmitted
		}

		/*
		 * Free up the Transmit Buffers. 
		 */
		fdbd->FD.FDStat = 0;
		fdbd->FD.FDSystem = 0;
#ifdef	SKB_RECYCLE
		msp_recycle_skb(skb);
#else
		dev_kfree_skb_any(skb);
#endif	/* SKB_RECYCLE */
		wakequeue = 1;
	} while (! MACOwnsTxFD(priv->TxFrmPtrFreeBuf));

	if (unlikely(wakequeue && netif_queue_stopped(priv->netdevice)))
	{
		/* 
		 * We freed a packet, let non-interrupt processing 
		 * free TX packets
		 */
		MSPEthTxIntDisable(priv);
		priv->netdevice->trans_start = jiffies;
		netif_wake_queue(priv->netdevice);
	}

	spin_unlock_irqrestore(&priv->lock, flags);
}


static inline void	MSPEthFreeTxBuffs(MSP_DEVICE *priv)
{
	if (TxFDInUse(priv->TxFrmPtrFreeBuf) && 
		! MACOwnsTxFD(priv->TxFrmPtrFreeBuf))
	{
		MSPEthFreeTxBuffsInternal(priv);
	}
}


/**********************************************************************
 * MSPEthMemoryFree
 *
 * Purpose:	Stop transmission and reception and free memory 
 *		that was used for buffers.
 *   
 * Parameters:	None
 *
 * Returns:	None    
 */

static void	MSPEthMemoryFree(MSP_DEVICE *pDrvCtrl) 
{
	int	Counter;
	unsigned long	flags;

	save_and_cli(flags);

	/* Prevent 1 second timer from accessing this structure */

	gEthDrvCtrl[pDrvCtrl->unit] = NULL;

	/* Prevent any calls from rest of kernel getting in.
	   Every routine that could be called must check priv != NULL */

	pDrvCtrl->netdevice->priv = NULL;

	/* Stop interrupts and disable rx and tx */

	MSPEthIntDisable(pDrvCtrl);
	MSPEthEnableTxHalt(pDrvCtrl);
	MSPEthDisableRx(pDrvCtrl);

	restore_flags(flags);

	/* Free up the Rx Buffers */

	if (pDrvCtrl->rskbp != NULL) 
	{
		for (Counter = 0; Counter < pDrvCtrl->rxBdCount; Counter++) 
		{
			if (pDrvCtrl->rskbp[Counter] != NULL)
			{
				dev_kfree_skb(pDrvCtrl->rskbp[Counter]);
			}
		}
		kfree(pDrvCtrl->rskbp);
	}

	/* Mark all tx descriptors not in use so we can free them */

	if (pDrvCtrl->TxFrmPtr != NULL)
	{
		FDBD	*fdbd;

		fdbd = pDrvCtrl->TxFrmPtr;

		do {
			if (TxFDInUse(fdbd) && MACOwnsTxFD(fdbd))
			{
				/* Force packet owned by us */

				fdbd->FD.FDCtl &= ~MAC_OWNS_FD;
				fdbd->FD.FDStat |= TX_STAT_Comp;
			}

			fdbd = (FDBD *) fdbd->FD.FDNext;

		} while (pDrvCtrl->TxFrmPtr != fdbd);

		MSPEthFreeTxBuffs(pDrvCtrl);
	}

	if (pDrvCtrl->BLFrmPtr_Orig)
		kfreeBDS(pDrvCtrl->BLFrmPtr_Orig);

	if (pDrvCtrl->FDAPtr_Orig)
		kfreeBDS(pDrvCtrl->FDAPtr_Orig);

	if (pDrvCtrl->TxFrmPtr_Orig)
		kfreeBDS(pDrvCtrl->TxFrmPtr_Orig);

#ifndef CONFIG_BRECIS_ENET_SCRATCH
	kfree(pDrvCtrl);
#endif	/* !CONFIG_BRECIS_ENET_SCRATCH */
}


#if 0
static void	MSPEthDisplayPhyInfo(int interfaceNum)
{
	MSP_DEVICE *pDrvCtrl;

	if (interfaceNum >= MSP_ETH_MAX_UNITS || gEthDrvCtrl[interfaceNum] == NULL)
	{
		printk(KERN_WARNING "Interface %d out of range\n", 
			 interfaceNum);
		return;
	}
	
	pDrvCtrl = gEthDrvCtrl[interfaceNum];
	
	switch (pDrvCtrl->phyType)
	{
	case PHYTYPE_INTEL971a:
		printk(KERN_INFO "Phy is %s\n", PHY_INTEL971a);
		break;
	case PHYTYPE_KENDINSWITCH:
		printk(KERN_INFO "Phy is %s\n", PHY_KENDINSWITCH);
		break;
	case PHYTYPE_KENDIN_8721:
		printk(KERN_INFO "Phy is %s\n", PHY_KENDIN_8721);
		break;
	case PHYTYPE_GENERIC:
		printk(KERN_INFO "Phy is %s\n", PHY_GENERIC);
		break;
	case PHYTYPE_NONE:
		printk(KERN_INFO "Phy is %s\n", PHY_NONE);
		break;
	}

	switch (pDrvCtrl->ethMode)
	{
	case ETH_AUTO:
		printk(KERN_INFO "Phy is set to AUTO NEGOTIATION\n");
		printk(KERN_INFO "Current Speed is %s\n", 
			(pDrvCtrl->speed == MSPEth_10MBS ? "10Mbps" : "100Mbps"));
		printk(KERN_INFO "Current Duplex is %s\n", 
			(pDrvCtrl->duplexMode == MAC_DPX_MODE_HALF ? "Half" : "Full"));
		break;
	case ETH_10Mbps_HALF:
		printk(KERN_INFO "Phy is set to 10 Mbps - Half Duplex\n");
		break;
	case ETH_10Mbps_FULL:
		printk(KERN_INFO "Phy is set to 10 Mbps - Full Duplex\n");
		break;
	case ETH_100Mbps_HALF:
		printk(KERN_INFO "Phy is set to 100 Mbps - Half Duplex\n");
		break;
	case ETH_100Mbps_FULL:
		printk(KERN_INFO "Phy is set to 100 Mbps - Full Duplex\n");
		break;
	}
}


static void MSPEthDisplayConfigInfo(int interfaceNum)
{
	MSP_DEVICE	*pDrvCtrl;

	if (interfaceNum >= MSP_ETH_MAX_UNITS || gEthDrvCtrl[interfaceNum] == NULL)
	{
		printk(KERN_WARNING "Interface %d out of range\n", 
			 interfaceNum);
		return;
	}
	
	pDrvCtrl = gEthDrvCtrl[interfaceNum];

	printk(KERN_INFO "\nReading MAC Registers for MAC # %d", 
		interfaceNum);		

	printk(KERN_INFO "\n\n");
	printk(KERN_INFO "------------------------------------------------\n");
	printk(KERN_INFO "Configuration Information for MAC: %d\n", 
		interfaceNum);
	printk(KERN_INFO "rxBdCount       = %d\n", pDrvCtrl->rxBdCount);
	printk(KERN_INFO "txBdCount       = %d\n", pDrvCtrl->txBdCount);
	printk(KERN_INFO "FDAPtr          = 0x%x\n", 
		(UINT) pDrvCtrl->FDAPtr);
	printk(KERN_INFO "FDAPtrCur       = 0x%x\n", 
		(UINT) pDrvCtrl->FDAPtrCur);
	printk(KERN_INFO "BLFrmPtr        = 0x%x\n", 
		(UINT) pDrvCtrl->BLFrmPtr);
	printk(KERN_INFO "TxFrmPtr        = 0x%x\n", 
		(UINT) pDrvCtrl->TxFrmPtr);
	printk(KERN_INFO "TxFrmPtrCur     = 0x%x\n", 
		(UINT) pDrvCtrl->TxFrmPtrCur);
	printk(KERN_INFO "TxFrmPtrFreeBuf = 0x%x\n", 
		(UINT) pDrvCtrl->TxFrmPtrFreeBuf);
	printk(KERN_INFO "\nVLAN Information\n");
	printk(KERN_INFO "----------------\n");

	printk(KERN_INFO "-------------------------------------------------\n");
}


static void MSPEthDisplayEthTxRxDescriptors(int interfaceNum)
{
	MSP_DEVICE *pDrvCtrl;

	if (interfaceNum >= MSP_ETH_MAX_UNITS || gEthDrvCtrl[interfaceNum] == NULL)
	{
		printk(KERN_WARNING "Interface %d out of range\n", 
			 interfaceNum);
		return;
	}

	pDrvCtrl = gEthDrvCtrl[interfaceNum];

	printk(KERN_INFO "\n\n");
	printk(KERN_INFO "-----------------------------------------------\n");
	printk(KERN_INFO "Tx, Rx Descriptor Info for MAC: %d\n", 
		interfaceNum);
	printk(KERN_INFO "FDAPtr[%d]          = 0x%08X\n", 
		interfaceNum, (UINT) pDrvCtrl->FDAPtr);
	printk(KERN_INFO "FDAPtrCur[%d]       = 0x%08X\n", 
		interfaceNum, (UINT) pDrvCtrl->FDAPtrCur);
	printk(KERN_INFO "BLFrmPtr[%d]        = 0x%08X\n", 
		interfaceNum, (UINT) pDrvCtrl->BLFrmPtr);
	printk(KERN_INFO "TxFrmPtr[%d]        = 0x%08X\n", 
		interfaceNum, (UINT) pDrvCtrl->TxFrmPtr);
	printk(KERN_INFO "TxFrmPtrCur[%d]     = 0x%08X\n", 
		interfaceNum, (UINT) pDrvCtrl->TxFrmPtrCur);
	printk(KERN_INFO "TxFrmPtrFreeBuf[%d] = 0x%08X\n", 
		interfaceNum, (UINT)pDrvCtrl->TxFrmPtrFreeBuf);

	printk(KERN_INFO "-----------------------------------------------\n");
}
#endif /* 0 */


#ifdef RESERVE2
#define DMA_SET_NORMAL \
	(DMA_CTL_M66EnStat | \
	 DMA_CTL_RxBigE | \
	 DMA_CTL_TxBigE | \
	 DMA_CTL_RxAlignSkip2)
#else
#define DMA_SET_NORMAL \
	(DMA_CTL_M66EnStat | \
	 DMA_CTL_RxBigE | \
	 DMA_CTL_TxBigE)
#endif

#define DMA_SET_TEST	(DMA_SET_NORMAL | DMA_CTL_TestMode)

static void	MacSetupRegistersTx(MSP_DEVICE *pDrvCtrl)
{
	WriteMacRegister(pDrvCtrl, DMA_CTL, DMA_SET_NORMAL | 64);

	WriteMacRegister(pDrvCtrl, MAC_CTL, 0x00000000);  
#if 0
	WriteMacRegister(pDrvCtrl, TXTHRSH, 0x00000200); 
#else
	WriteMacRegister(pDrvCtrl, TXTHRSH, 0x000005EA); 
#endif
/*	WriteMacRegister(pDrvCtrl, TXTHRSH, 0x00000040); */
	WriteMacRegister(pDrvCtrl, TXPOLLCTR, 0x0000FFFF);
	WriteMacRegister(pDrvCtrl, TX_CTL,  0x00000000);	/* Turn off Tx */
}


static void MacSetupRegistersRx(MSP_DEVICE *pDrvCtrl, int fdaSize)
{
	PRINTK(KERN_INFO "FDASize = %d, FDAPTR %x, FDA_LIM %d, BLFRMPTR %x\n",
		(int) fdaSize, 
		(unsigned int) pDrvCtrl->FDAPtr,
		(int) fdaSize - 32 * sizeof(BD),
		(unsigned int) pDrvCtrl->BLFrmPtr);

	WriteMacRegister(pDrvCtrl, FDA_BAS, (u32) pDrvCtrl->FDAPtr); 

	/* Leave Room for 1 FD and 30 BD's */
	WriteMacRegister(pDrvCtrl, FDA_LIM, (fdaSize - 32 * sizeof(BD)));  

	WriteMacRegister(pDrvCtrl, RXFRAGSIZE, 0x00); 

	WriteMacRegister(pDrvCtrl, ARC_CTL, 0);
	WriteMacRegister(pDrvCtrl, ARC_ENA, 0x00);  /* Disable ARC Entries */

	WriteMacRegister(pDrvCtrl, BLFRMPTR, (u32)pDrvCtrl->BLFrmPtr); 
	WriteMacRegister(pDrvCtrl, RX_CTL, 0x00000000);	/* Turn off Rx */
}


static int getEthUnit(struct net_device *dev)
{
	int	unit;
	
	/* 
	 * All our devices are eth%d devices (please see space.c).
	 * Get the unit number from the %d as our unit number.
	 */
	if (strncmp("eth", dev->name, strlen("eth")) != 0)
		return -1;		/* not eth%d device */

	if (sscanf(&dev->name[strlen("eth")], "%d", &unit) != 1)
		return -1;

	if (unit < 0 || unit >= MSP_ETH_MAX_UNITS)
		return -1;

	return unit;
}


int compute_hwunit(int unit)
{
	int logicalunit = 0;
	int hwunit;
	int noexist = MSP_ETH_MAX_UNITS;

	for (hwunit = 0; hwunit < MSP_ETH_MAX_UNITS; hwunit++)
	{
		if (identify_enet(hwunit) != FEATURE_NOEXIST)
		{
			if (logicalunit++ == unit)
				return hwunit;
		}
		else
			noexist = hwunit;
	}
	return noexist;
}


#ifdef	CONFIG_BRECIS_ENET_SCRATCH
MSP_DEVICE	DrvCtrl[MSP_ETH_MAX_UNITS] __fast = { {0} };
#endif	/* CONFIG_BRECIS_ENET_SCRATCH */

static int	MSPEthSetup(struct net_device *dev)
{
	int		ethspeedduplex;
	int		ethswitch;
	int		unit;
	int		hwunit;
	MSP_DEVICE	*pDrvCtrl;

	/* Free any old MSP_DEVICE structure */

	pDrvCtrl = dev->priv;

	if (pDrvCtrl != NULL)
	{
		MSPEthReset(pDrvCtrl);
		MSPEthMemoryFree(pDrvCtrl);
	}

	unit = getEthUnit(dev);
	if (unit < 0 || gEthDrvCtrl[unit] != NULL || unit >= MSP_ETH_MAX_UNITS)
	{
		printk(KERN_ERR "%s: Bad Unit number or already in use\n",
			dev->name);
		return MSP_FAIL;
	}

	/* Create new MSP_DEVICE structure */
#ifdef	CONFIG_BRECIS_ENET_SCRATCH
	dev->priv = &DrvCtrl[unit];
#else
	dev->priv = kmalloc(sizeof(MSP_DEVICE), GFP_KERNEL);
#endif	/* CONFIG_BRECIS_ENET_SCRATCH */

	if (dev->priv == NULL) {
		printk(KERN_ERR "%s: MSPEthSetup: kmalloc returned NULL\n",
			dev->name);
		return MSP_FAIL;
	}

	memset(dev->priv, 0, sizeof(MSP_DEVICE));

	pDrvCtrl = dev->priv;
	pDrvCtrl->netdevice = dev;
	pDrvCtrl->rx_bh = &rx_bh[unit];
	pDrvCtrl->reload_tq = &reload_tq_g[unit];
	if (!unit_init[unit]) {
		INIT_TQUEUE(pDrvCtrl->rx_bh, RXbh, dev);
		INIT_TQUEUE(pDrvCtrl->reload_tq, reload_tq, dev);
		unit_init[unit] = 1;
	}

#ifdef ETHERNETUSESCRATCHRAM
	if (identify_spad() == FEATURE_NOEXIST)
	{
		pDrvCtrl->txBdCount = ETH_TX_BD_COUNT;
		pDrvCtrl->rxBdCount = ETH_RX_BD_COUNT;
	} else {
		pDrvCtrl->txBdCount = ETH_TX_BD_COUNT_SPAD;
		pDrvCtrl->rxBdCount = ETH_RX_BD_COUNT_SPAD;
	}
#else
	pDrvCtrl->txBdCount = ETH_TX_BD_COUNT;
	pDrvCtrl->rxBdCount = ETH_RX_BD_COUNT;
#endif
	hwunit = compute_hwunit(unit);
	pDrvCtrl->enetType = identify_enet(hwunit);
	pDrvCtrl->enetTxD = identify_enetTxD(hwunit);

	pDrvCtrl->unit = unit;

	pDrvCtrl->mac_base = (MAC_BASE *)MacBases[hwunit];

	pDrvCtrl->phy_base = &PhyBases[phyAddresses[unit] >> 16];

	if (hwunit == 2)
	{
		switch (identify_family())
		{
		case FAMILY_POLO:
		case FAMILY_ZEUS:
			/* If Polo or Zeus, set GPIO bit to enable MAC 2 */
			*GPIO_CFG7_REG |= POLO_MAC2_ENABLE;
			break;

		default:
			break;
		}
	}

	pDrvCtrl->rst_base = MacRSTBases[hwunit];

	ethspeedduplex = gethspeedduplex[unit];
	ethswitch = gethswitch[unit];

	switch (ethspeedduplex)
	{
	case 10:
		pDrvCtrl->ethMode = ETH_10Mbps_HALF;
		pDrvCtrl->speed = MSPEth_10MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_HALF;
		break;
	case 20:
		pDrvCtrl->ethMode = ETH_10Mbps_FULL;
		pDrvCtrl->speed = MSPEth_10MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_FULL;
		break;
	case 100:
		pDrvCtrl->ethMode = ETH_100Mbps_HALF;
		pDrvCtrl->speed = MSPEth_100MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_HALF;
		break;
	case 200:
		pDrvCtrl->ethMode = ETH_100Mbps_FULL;
		pDrvCtrl->speed = MSPEth_100MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_FULL;
		break;
	case -1:
	default:
		pDrvCtrl->ethMode = ETH_AUTO;
	}

	if (ethswitch)
		pDrvCtrl->phyType = PHYTYPE_NONE;
	else
		pDrvCtrl->phyType = PHYTYPE_GENERIC;

	gEthDrvCtrl[unit] = pDrvCtrl;

	PRINTK(KERN_INFO "%s: hwunit %d, mac_base %x\n", 
		dev->name, hwunit, (unsigned int) pDrvCtrl->mac_base);

	MSPEthReset(pDrvCtrl);

	MSPEthIntDisable(pDrvCtrl);

	if (MacSetupTxBDs(pDrvCtrl) != MSP_SUCCESS)
	{
		MSPEthMemoryFree(pDrvCtrl);
		return MSP_FAIL;
	}

	if (MacSetupRxBDs(pDrvCtrl) != MSP_SUCCESS)
	{
		MSPEthMemoryFree(pDrvCtrl);
		return MSP_FAIL;
	}

	MacSetupRegistersTx(pDrvCtrl);
	MacSetUpTxFrmPtrRegister(pDrvCtrl, (u32)pDrvCtrl->TxFrmPtrCur);

#if defined(CONFIG_VLAN_8021Q) || defined(CONFIG_VLAN_8021Q_MODULE)
	MSPEthEnableVLAN(pDrvCtrl);
#endif

	if (mspeth_inited)
		MSPEthPhyInit(pDrvCtrl);

	return MSP_SUCCESS;
}


/*************************************************************************
 * ethPhyInit - initialize and configure the PHY device.
 *
 * RETURNS: OK, or ERROR
 * Needs to be modified according to the PHY, and 
 * also there are two different PHYs connected to MAC 1 and 2.....
 * needs to be modified according to the SPEC of Level1 971 and 905.
 */

static void determinePhyAddress(MSP_DEVICE *pDrvCtrl)
{
	u32	j;
	u32	ethIndex;
	u32	hwIndex;
	u32	myPhyAddress;
	u32	phyAddress;
	u32	phyIndex;
	U32	reg1;

	myPhyAddress = phyAddresses[pDrvCtrl->unit];
	hwIndex = myPhyAddress >> 16;
	phyIndex = myPhyAddress & 0xffff;

	if (phyIndex != MD_DYNAMIC_PHY)
	{
		pDrvCtrl->macPhyId = phyIndex << MD_CA_PhyShift;
		printk("eth%d: PHY is statically configured %d:%d\n",
		       pDrvCtrl->unit,
		       (unsigned int) hwIndex, (unsigned int) phyIndex);
		return;
	}
	
	/*
	 * determine PHY address
	 * hwIndex is index for selecting mdc/mdio hardware
	 * phyIndex is phy address index
	 */
	for (phyIndex = 0, ethIndex = 0; phyIndex < MD_MAX_PHY; phyIndex++)
	{
		/* We are testing for the phyAddress hwIndex:phyIndex */
		phyAddress = (hwIndex << 16) | phyIndex;

		for (j = 0; j < MSP_ETH_MAX_UNITS; j++)
			if (phyAddress == phyAddresses[j])
				break;

		/* Check for phyAddress under test statically assigned */
		if (j < MSP_ETH_MAX_UNITS)
			continue;	/* already statically assigned */

		pDrvCtrl->macPhyId = phyIndex << MD_CA_PhyShift;

		pDrvCtrl->phyprobe = 1;
		reg1 = readPhyRegister(pDrvCtrl, PHYG_STATUS_REG_ONE);
		pDrvCtrl->phyprobe = 0;

		if (reg1 != 0 && reg1 != 0xffff && (reg1 & PHYGb_MEDIAMASK) != 0)
		{
			/* 
			 * We found a PHY we can dynamically assign.
			 * We assign PHYs matching those ethernet interfaces
			 * that take dynamic PHY assignment.  I.e., the first
			 * dynamically assignable PHY goes with the first
			 * ethernet interface that wants a PHY dynamically
			 * assigned on that particular hardware mdc/mdio.
			 * ethIndex is the index for the current ethernet
			 * interface to check for dynamic assignment.
			 * When ethIndex equals pDrvCtrl->unit, we found
			 * our interface.
			 */
			while (phyAddresses[ethIndex] != myPhyAddress)
				ethIndex++;

			if (ethIndex == pDrvCtrl->unit)
				break;	/* We found PHY for our interface */

			/* Increment to next possible ethernet interface */
			ethIndex++;
		}
	}

	if (phyIndex == MD_MAX_PHY) 
	{
		printk(KERN_INFO "eth%d: no PHY found, using default\n",
			pDrvCtrl->unit);
		pDrvCtrl->macPhyId = 0 << MD_CA_PhyShift;
	}
}


static void MSPEthPhyInit(MSP_DEVICE *pDrvCtrl)
{
	switch (pDrvCtrl->phyType)
	{
	case PHYTYPE_NONE:
	case PHYTYPE_KENDINSWITCH:
		MSPEthPhyInit_complete(pDrvCtrl);
		brecis_eth_complete_open(pDrvCtrl);
		break;

	case PHYTYPE_GENERIC:
	case PHYTYPE_KENDIN_8721:
	case PHYTYPE_INTEL971a:
		determinePhyAddress(pDrvCtrl);
		MSPPHYGSetup(pDrvCtrl);
		break;
	}
}


static void MSPEthPhyInit_complete(MSP_DEVICE *pDrvCtrl)
{
	struct net_device	*dev = pDrvCtrl->netdevice;

	MSPEthSetDuplex(pDrvCtrl, pDrvCtrl->duplexMode);

	switch (pDrvCtrl->enetTxD)
	{
	case ENETTXD_RISING:
		/* Set TxD rising edge - I was told this was best default */

		MSPEthRMIITxDRisingEdge(pDrvCtrl);
		break;

	case ENETTXD_FALLING:
		/* Set TxD falling edge */

		MSPEthRMIITxDFallingEdge(pDrvCtrl);
		break;

	default:
		/* Do not touch the TxD bit */
		break;
	}

	
	if (brctl_present)
	{
		/* Set speed of RMII based on PHY setup/negotiation */

		if (pDrvCtrl->speed == MSPEth_10MBS)
			MSPEthRMIIEnable10Mbit(pDrvCtrl);
		else
			MSPEthRMIIEnable100Mbit(pDrvCtrl);
	}

	if (pDrvCtrl->phyType == PHYTYPE_KENDINSWITCH
		|| pDrvCtrl->phyType == PHYTYPE_NONE)
	{
		/* A SWITCH ALWAYS has carrier */
		printk(KERN_INFO "eth%d: Switch is set for %d Mbps %s Duplex\n",
		       pDrvCtrl->unit,
		       pDrvCtrl->speed == MSPEth_10MBS ? 10 : 
		       (pDrvCtrl->speed == MSPEth_100MBS ? 100 : 0),
		       pDrvCtrl->duplexMode == MAC_DPX_MODE_HALF ?
		       "Half" : 
		       (pDrvCtrl->duplexMode == MAC_DPX_MODE_FULL ? 
			"Full" : "UNKNOWN"));
		return;
	}

	/*
	 * Tell upper layer our carrier status
	 */
	if (MSPPHYGCheckCable(pDrvCtrl) != 0)
	{
		/* If link is up */

		if (! netif_carrier_ok(dev))
		{
			netif_carrier_on(dev);
		}
	}
	else
	{
		/* If link is not up */

		if (netif_carrier_ok(dev))
		{
			netif_carrier_off(dev);
		}
	}
} 


/*************************************************************************
 * MSPEthSetDuplex - sets the duplex mode on the MAC
 *
 * NOTE: This function is called after setting the PHY and the
 * autonegotiatin is complete on the PHY. Find out the duplex
 * mode from the PHY and set it at the MAC to match it.
 *
 * Input: duplexMode:
 *		0 - Half Fuplex
 *		1 - Full Duplex
 */

static void	MSPEthSetDuplex(MSP_DEVICE *pDrvCtrl, int duplexMode)
{
	if (duplexMode == MAC_DPX_MODE_HALF)
	{
		AndMacReg(pDrvCtrl, MAC_CTL, ~MAC_CTL_FullDup);
	}
	else if (duplexMode == MAC_DPX_MODE_FULL) 
	{
		OrMacReg(pDrvCtrl, MAC_CTL, MAC_CTL_FullDup);
	}
} 


#ifdef DEBUGDISPLAYMACREGISTERS
static void MSPEthDisplayMacRegisters(int interfaceNum)
{
	MSP_DEVICE *pDrvCtrl;

	if (interfaceNum >= MSP_ETH_MAX_UNITS || gEthDrvCtrl[interfaceNum] == NULL)
	{
		printk(KERN_WARNING "Interface %d out of range\n", interfaceNum);
		return;
	}
	
	pDrvCtrl = gEthDrvCtrl[interfaceNum];

	printk(KERN_INFO "Reading MAC Registers for MAC #  %d\n", interfaceNum);		
	
	printk(KERN_INFO 
		"Reading MAC Register DMA_CTL      0x00\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, DMA_CTL));
	
	printk(KERN_INFO
		"Reading MAC Register TXTHRSH      0x08\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, TXTHRSH));

	printk(KERN_INFO
		"Reading MAC Register TXPOLLCTR    0x0c\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, TXPOLLCTR));

	printk(KERN_INFO
		"Reading MAC Register BLFRMPTR     0x10\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, BLFRMPTR));

	printk(KERN_INFO 
		"Reading MAC Register RXFRAGSIZE   0x14\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, RXFRAGSIZE));

	printk(KERN_INFO
		"Reading MAC Register INT_EN       0x18\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, INT_EN));
	
	printk(KERN_INFO
		"Reading MAC Register FDA_BAS      0x1c\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, FDA_BAS));

	printk(KERN_INFO
		"Reading MAC Register FDA_LIM      0x20\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, FDA_LIM));

	printk(KERN_INFO
		"Reading MAC Register INT_SRC      0x24\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, INT_SRC));

	printk(KERN_INFO
		"Reading MAC Register PAUSECNT     0x30\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, PAUSECNT));

	printk(KERN_INFO
		"Reading MAC Register REMPAUCNT    0x34\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, REMPAUCNT));

	printk(KERN_INFO
		"Reading MAC Register TXCONFRMSTAT 0x38\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, TXCONFRMSTAT));

	printk(KERN_INFO
		"Reading MAC Register MAC_CTL      0x40\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, MAC_CTL));

	printk(KERN_INFO
		"Reading MAC Register ARC_CTL      0x44\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, ARC_CTL));

	printk(KERN_INFO
		"Reading MAC Register TX_CTL       0x48\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, TX_CTL));

	printk(KERN_INFO
		"Reading MAC Register TX_STAT      0x4c\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, TX_STAT));

	printk(KERN_INFO
		"Reading MAC Register RX_CTL       0x50\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, RX_CTL));

	printk(KERN_INFO
		"Reading MAC Register RX_STAT      0x54\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, RX_STAT));
	
	printk(KERN_INFO 
		"Reading MAC Register MD_DATA      0x58\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, MD_DATA));
	
	printk(KERN_INFO
		"Reading MAC Register MD_CA        0x5c\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, MD_CA));

	printk(KERN_INFO
		"Reading MAC Register ARC_ADR      0x60\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, ARC_ADR));

	printk(KERN_INFO 
		"Reading MAC Register ARC_DATA     0x64\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, ARC_DATA));
	
	printk(KERN_INFO
		"Reading MAC Register ARC_ENA      0x68\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, ARC_ENA));
	
	printk(KERN_INFO
		"Reading MAC Register MISS_CNT     0x7c\t 0x%08X\n",
		(unsigned int) ReadMacRegister(pDrvCtrl, MISS_CNT));

	switch (identify_family())
	{
	case FAMILY_POLO:
	case FAMILY_ZEUS:
		printk(KERN_INFO
		       "Reading MAC Register MAC_BRCTRL_REG 0x8c\t 0x%08X\n\n",
		       (unsigned int) ReadMacRegister(pDrvCtrl, 
						      MAC_BRCTRL_REG));
		break;

	default:
		break;
	}
}
#endif


/************************************************************************
 * Check for a network adapter of this type, and return '0' iff one exists.
 *  - we just allocate them the number of times I'm called
 */

int __init 
brecis_eth_probe(struct net_device *dev)
{
	int	unit;	
	int	hwunit;
	u32	baseAddr;
	char	enetType;
	unsigned char	ptr[20];

	unit = getEthUnit(dev);
	// check if all allocated

	IDBG(KERN_INFO "getEthUnit returns %d for %s\n", unit, dev->name);

	hwunit = compute_hwunit(unit);
	if (unit < 0 || gEthDrvCtrl[unit] != NULL || hwunit >= MSP_ETH_MAX_UNITS)
		return -ENODEV;

	enetType = identify_enet(hwunit);

	if (enetType == FEATURE_NOEXIST)
		return -ENODEV;	/* interface doesn't exist for this chip */

	/* Retrieve (from environment) and print the ethernet address. */
	if (get_ethernet_addr(MacName[unit], ptr))
	{ // no ethernet MAC address in environment
		pkinit(KERN_INFO " No Mac addr specified for eth%d %s\n",
			unit, MacName[unit]);
		return -ENODEV;
	}

	if (*ptr & 1)
	{
		pkinit(KERN_INFO "Bad Multicast Mac addr specified for eth%d "
			"%s %02x:%02x:%02x:%02x:%02x%02x\n",
			unit, MacName[unit], 
			ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5]);
		return -ENODEV;
	}

	IDBG(KERN_INFO "Retrieved MAC address for eth%d %s, enetType %c\n",
		unit, MacName[unit], enetType);

	// SET_MODULE_OWNER(dev);
	dev->irq = MacIRQs[hwunit];
	baseAddr = MacBases[hwunit];

	if (check_mem_region(baseAddr, MSPEthExtent))
		return -ENODEV;

#ifndef	MODULE
	{
		static u8 __initdata	printed_version;

		if (!printed_version++)
			pkinit(KERN_INFO "%s", version);
	}
#endif	/* !MODULE */

	if (brecis_eth_probe1(dev, baseAddr, unit, ptr) == 0)
	{
		return 0;
	}
	return -ENODEV;
}


/************************************************************************
 * This routine initializes all the vars in dev
 */

static void	brecis_eth_InitVars(struct net_device *dev, int ioaddr)
{
	dev->base_addr	= ioaddr;

	dev->open		= &brecis_eth_open;
	dev->stop		= &brecis_eth_close;
	dev->hard_start_xmit	= &brecis_eth_send_packet;
	dev->get_stats	= &brecis_eth_get_stats;
	dev->set_multicast_list = &set_multicast_list;
	dev->tx_timeout	= &brecis_eth_tx_timeout;
	dev->watchdog_timeo	= MY_TX_TIMEOUT; 
#if 1
	dev->do_ioctl		= &brecis_eth_do_ioctl;
#endif
	dev->set_mac_address	= &brecis_eth_set_mac_address;

#ifdef	CONFIG_NET_FASTROUTE
	dev->accept_fastpath = msp_accept_fastpath;
#endif	/* CONFIG_NET_FASTROUTE */
#ifdef	SKB_RECYCLE
	if (msp_recycle_list.next == NULL)
		skb_queue_head_init(&msp_recycle_list);
#endif	/* SKB_RECYCLE */
}


/************************************************************************
 * This is the real probe routine. Linux has a history of friendly device
 * probes on the ISA bus. A good device probes avoids doing writes, and
 * verifies that the correct device exists and functions.
 */

static int __init 
brecis_eth_probe1(struct net_device *dev, int ioaddr, int unit, unsigned char *ptr)
{
	int	i;	
	int	irqval;
	MSP_DEVICE	*pDrvCtrl;

	IDBG(KERN_INFO "%s: found at %#3x\n", dev->name, ioaddr);
	/* Fill in the 'dev' fields. */
	dev->base_addr = ioaddr;

	dev->addr_len = EADDR_LEN;
	for (i = 0; i < dev->addr_len; i++)
		dev->dev_addr[i] = ptr[i];

	IDBG(KERN_INFO "%s: MAC address is %02x:%02x:%02x:%02x:%02x:%02x\n",
		dev->name, ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5]);

	/*
	 * The ethernet MACS have fixed interrupts, allocate the interrupt
	 * vector now. There is no point in waiting since no other device
	 * can use the interrupt, and this marks the irq as busy. Jumpered
	 * interrupts are typically not reported by the boards. MAC 0 is 
	 * fixed at IRQ 10, MAC 1 is fixed IRQ11
	 */
	irqval = request_irq(dev->irq, &brecis_eth_interrupt,
			SA_SAMPLE_RANDOM, cardname, dev);
	if (irqval)
	{
		pkinit(KERN_WARNING "%s: unable to get IRQ %d (irqval=%d).\n",
			dev->name, dev->irq, irqval);
		return -EAGAIN;
	}

	/* Grab the region so that no one else tries to probe our ioports. */
	if (! request_mem_region(ioaddr, MSPEthExtent, cardname))
	{
		pkinit(KERN_WARNING
			"%s: unable to get memory/io address region %x\n",
			dev->name, ioaddr);
		return -EAGAIN;
	}
	ether_setup(dev);
	if (MSPEthSetup(dev) != MSP_SUCCESS)
		return -EAGAIN;

	pDrvCtrl = dev->priv;

	brecis_eth_InitVars(dev, ioaddr);
	return 0;
}


/***********************************************************************
 * Transmit timeout
 *	Kernel calls this routine when a transmit has timed out
 */

static void 
brecis_eth_tx_timeout(struct net_device *dev)
{
#if 0
	int	stp;
	FDBD	*ofree;
	static int	ctr;

	stp = netif_queue_stopped(priv->netdevice);
	ofree = priv->TxFrmPtrFreeBuf;
  
	if (ctr++ < 10)
	{
		printk("TxFrmPtrFreeBuf descriptor:\n");
		printk("   Ethernet Unit: %d\n", priv->unit);
		printk("   Address: %p\n", ofree);
		printk("      Stat: %x  Ctl: %x\n",
			(unsigned int)ofree->FD.FDStat, ofree->FD.FDCtl);
	}
#endif	  

	if (dev->priv == 0)
		return;

	setup_delayed_reload(dev);

#if 0
	printk("brecis_eth_tx_timeout called txintr %x", priv->dbgint);
	if (stp != netif_queue_stopped(priv->netdevice))
		printk(" queue_stop changed from %x to %x ", stp,
			netif_queue_stopped(priv->netdevice));
	if (ofree != priv->TxFrmPtrFreeBuf)
		printk(" free ptr changed from %p to %p", ofree,
			priv->TxFrmPtrFreeBuf);
	printk("\n");
#endif
}


#ifdef ETHERNETUSESCRATCHRAM
static u8	*kmallocSRAM(int size)
{
	int index;
	int tmp;
	int tmpsize = ((size + 0x0f) & ~0x0f);

	if (identify_spad() == FEATURE_NOEXIST)
		return NULL;

	for (index = 0; index < MAXSCRATCHRAMPARTITIONS; index++)
	{
		if (scratchramstate[index] != SCRATCH_FREE)
			continue;

		if (scratchramaddr[index+1] - scratchramaddr[index] >= tmpsize)
		{
			break;
		}
	}

	if (scratchramstate[index] != SCRATCH_FREE)
		return NULL;

	scratchramstate[index] = SCRATCH_INUSE;
	
	if (scratchramaddr[index+1] - scratchramaddr[index] == tmpsize)
		return (UINT8 *) scratchramaddr[index];

	if (scratchramstate[index+1] != SCRATCH_FREE)
	{
		for (tmp = MAXSCRATCHRAMPARTITIONS; tmp > index; tmp--)
		{
			scratchramaddr[tmp] = scratchramaddr[tmp-1];
			scratchramstate[tmp] = scratchramstate[tmp-1];
		}

		scratchramstate[index+1] = SCRATCH_FREE;
	}
	
	scratchramaddr[index+1] = scratchramaddr[index] + tmpsize;
	return (u8 *) scratchramaddr[index];
}


static void	kfreeSRAM(void *buf)
{
	int	index;
	unsigned long	addr = (unsigned long) buf;

	for (index = 0; index < MAXSCRATCHRAMPARTITIONS; index++)
	{
		if (scratchramstate[index] != SCRATCH_INUSE)
			continue;

		if (addr == scratchramaddr[index])
		{
			scratchramstate[index] = SCRATCH_FREE;

			if (scratchramstate[index+1] != SCRATCH_FREE)
				return;

			for (index++; index < MAXSCRATCHRAMPARTITIONS; index++)
			{
				scratchramaddr[index] =
					scratchramaddr[index+1];
				scratchramstate[index] =
					scratchramstate[index+1];
			}

			return;
		}
	}
}
#endif


static u8	*kmallocBDS(int size)
{
#ifdef ETHERNETUSESCRATCHRAM
	u8	*buf;

	buf = kmallocSRAM(size);

	if (buf != NULL)
		return buf;
#endif

	return kmalloc(size + 15, GFP_KERNEL | __GFP_DMA);
}


static void	kfreeBDS(void *buf)
{
#ifdef ETHERNETUSESCRATCHRAM
	if ((unsigned long) buf >= SCRATCHRAMADDR &&
	    (unsigned long) buf < (SCRATCHRAMADDR + ETHERNETSCRATCHRAMLEN))
	{
		kfreeSRAM(buf);
	}
	else
#endif
	{
		kfree(buf);
	}
}


static u8	*memalign16(u8 *buf)
{
	return (u8 *) (((unsigned long) buf + 15) & ~15);
}


#ifdef XCONFIG_NET_HW_FLOWCONTROL
static void	msp_xon(struct net_device *dev)
{
	MSP_DEVICE	*priv = dev->priv;

	printk(KERN_INFO "msp_xon called\n");
	clear_bit(priv->fc_bit, &netdev_fc_xoff);
	/* NOT READY! Needs port to MSP
	if (dev->start)
		outb_p(ei_local->rxcfg, e8390_base + EN0_RXCR);
	 */
}
#endif


#ifdef CONFIG_NET_FASTROUTE
static int __hotnet	msp_accept_fastpath(struct net_device *dev,
			struct dst_entry *dst)
{
	struct net_device	*odev = dst->dev;

	if (dst->ops->protocol != __constant_htons(ETH_P_IP) ||
		(odev->type != ARPHRD_ETHER || odev->accept_fastpath == NULL)) {
		tpk("N");
		return -1;
	}

	return 0;
}


/*
   Return values:

   0 - packet has gone by fast path.
   1 - packet was going to go by fast path, but output device is busy (semifast)
   2 - fast path is hit, but packet is a bit strange. (NI)
   3 - fast path miss.
   4 - oom.
 */

static int __hotnet	msp_fast_route(struct net_device *dev, struct sk_buff *skb,
			    int len, int entry)
{
	MSP_DEVICE	*priv = dev->priv;
	struct ethhdr	*eth = (void *)skb->data;
	struct rtable	*rt;

	if (unlikely(priv == 0))
		return 4;

	if (likely(eth->h_proto == __constant_htons(ETH_P_IP))) {
		struct iphdr	*iph;
		unsigned	h;

		iph = (struct iphdr *)(skb->data + ETH_HLEN);
		h = (iph->daddr ^ iph->saddr) & NETDEV_FASTROUTE_HMASK;
		pref(eth, 0x020);
		rt = (struct rtable *)(dev->fastpath[h]);
		tpk(rt ? "R" : "r");
		if (likely(rt && iph->daddr == rt->key.dst &&
			rt->u.dst.obsolete == 0)) {
			struct net_device	*odev = rt->u.dst.dev;

//			dev_fastroute_stat.hits++;
			tpk("M");

			if (unlikely(*(u8*)iph != 0x45 || (eth->h_dest[0] & 1) ||
				!neigh_is_valid(rt->u.dst.neighbour) ||
				iph->ttl <= 1))
					goto alas2;

			ip_decrease_ttl(iph);

			{
				struct sk_buff	*skb2 = getskb(priv);

				if (unlikely(skb2 == NULL))
					goto oom;

				priv->rskbp[entry] = skb2;
				skb2->dev = dev;
			}

			skb_put(skb, len);

//			ip_statistics.IpInReceives++;
//			ip_statistics.IpForwDatagrams++;

			/* Could use hh cache */
			memcpy(eth->h_source, odev->dev_addr, 6);
			memcpy(eth->h_dest, rt->u.dst.neighbour->ha, 6);
			skb->dev = odev;

#ifdef FAST_SKB_RECYCLE /* DO NOT DEFINE IT! READ COMMENT */
			/* We could use fast buffer recycling here if odev
			   is not DMAing.

			   The only problem is that we must allocate skb2
			   BEFORE we lose skb, otherwise we would make hole in
			   tulip rx array. Hence, to implement FAST_SKB_RECYCLE
			   we need always keep at least one skb in a safe place.
			 */
			atomic_inc(&skb->users);
#endif

#if 0
			/* If output device is busy try to poll it for a while... */
			{   
				long i;
				for(i=0; test_bit(0, (void*)&odev->tbusy) && i<50000 ;i++);
			}
#endif

			if (likely(/*odev->tx_semaphore &&
				odev->tbusy == 0 &&
				odev->interrupt == 0 &&*/
				odev->hard_start_xmit(skb, odev) == 0)) {
#ifdef FAST_SKB_RECYCLE
				if (likely(atomic_read(&skb->users) == 1)) {
					skb->tail = skb->data;
					skb->len = 0;
				}
#endif
				tpk("$");
//				dev_fastroute_stat.succeed++;
				return 0;
			}
#ifdef FAST_SKB_RECYCLE
			atomic_dec(&skb->users);
#endif

			/* Otherwise... */
			skb->pkt_type = PACKET_FASTROUTE;
			skb->nh.raw = skb->data + ETH_HLEN;
			skb->protocol = __constant_htons(ETH_P_IP);
//			dev_fastroute_stat.deferred++;
			tpk("1");
			return 1;
		}
	}
	tpk("4");
	return 4;

oom:
	tpk("3");
	return 3;

alas2:
#if	1	//def not_yet
	skb->dst = dst_clone(&rt->u.dst);
	tpk("2");
	return 2;
#else
	tpk("5");
	return 4;
#endif	/* not_yet */
}
#endif	/* CONFIG_NET_FASTROUTE */


/*
 * msp_process_rx_error - Process error indications on a packet
 *
 * This code is separated because it should not be done often and
 * should not be in the hot section.
 */

static void	msp_process_rx_error(MSP_DEVICE *pDrvCtrl, u32 status)
{
	int	errorFound = 0;

	pDrvCtrl->stats.rx_errors++;
	pDrvCtrl->stats.rx_dropped++;	// Do all count as a dropped pkt??
	if (status & (RCVD_LEN_ERR | RCVD_LONG_ERR)) {
		pDrvCtrl->stats.rx_length_errors++;
		errorFound = 1;
	}

	if (status & RCVD_OVRFLW_ERR) {
		pDrvCtrl->stats.rx_over_errors++;
		errorFound = 1;
	}
	else if (status & RCVD_CRC_ERR) {
		pDrvCtrl->stats.rx_crc_errors++;
		errorFound = 1;
	}

	if (status & RCVD_PAR_ERR) {
		pDrvCtrl->stats.rx_fifo_errors++;
		errorFound = 1;
	}

	if (status & RCVD_ALIGN_ERR) {
		pDrvCtrl->stats.rx_frame_errors++;
		errorFound = 1;
	}

	if (errorFound == 0)
		pDrvCtrl->stats.rx_missed_errors++;
}


/*
 * msp_rx_mult_bd - Process multiple received BDs.
 *
 * This code is separated out because it should never really be executed
 * and should not be in the hot section.
 */

static void	msp_rx_mult_bd(MSP_DEVICE *pDrvCtrl)
{
	BLFDBD	*bl;
	BD	*bd;
	FDBD	*pFDBD;
	int	bdNum;
	int	bdCount;

	pFDBD = (FDBD *) &(pDrvCtrl->FDAPtrCur->FD); 
	bdNum = pFDBD->FD.FDCtl & FDCTL_BD_COUNT_MASK;
	bl = pDrvCtrl->BLFrmPtr;

	for (bdCount = 0; bdCount < bdNum; bdCount++)
	{
		int	bdix;

		bdix = pFDBD->BD[bdCount].BDStat;
		bd = &bl->BD[bdix];
		bd->BuffData =
			getRxBuffData(pDrvCtrl->rskbp[bdix]->data);
		bd->BuffLength = MSP_END_BUFSIZE;
		bd->BDCtl      = MAC_OWNS_BD;

		++bdCount;
		if (bdCount < bdNum) {
			bdix = pFDBD->BD[bdCount].BDStat;
			bd = &bl->BD[bdix];
			bd->BuffData =
				getRxBuffData(pDrvCtrl->rskbp[bdix]->data);
			bd->BuffLength = MSP_END_BUFSIZE;
			bd->BDCtl      = MAC_OWNS_BD;
		}
		pFDBD->BD[bdCount].BDCtl = MAC_OWNS_BD;
	}
	pFDBD->FD.FDCtl = MAC_OWNS_FD;
}



/******************************************************************************
 * MSPEthReceive - Pass a received frame to the next layer up
 *
 * RETURNS: 0 if no more RX descriptors to process, non-zero otherwise
 */

static int __hotnet	MSPEthReceive(MSP_DEVICE *pDrvCtrl)
{
#ifdef DEBUGINTCOUNT
	int		count = 0;
#endif
	u32		len;
	int		maxloopcount;
	u32		status;
	int		bdNum;
	FDBD		*pFDBD;
	struct sk_buff	*skb;
#if 0
	FD		*tempFd;

	tempFd = (FD *) &(pDrvCtrl->FDAPtrCur->FD);
	PRINTK(KERN_INFO "Cur-%x Next-%x System-%x Stat-%x Ctl-%x Length-%x\n",
		   (unsigned int) tempFd,
		   (unsigned int) tempFd->FDNext,
		   (unsigned int) tempFd->FDSystem,
		   (unsigned int) tempFd->FDStat,
		   (unsigned int) tempFd->FDCtl,
		   (unsigned int) tempFd->FDLength);
#endif

	for (maxloopcount = pDrvCtrl->rxBdCount >> ETH_RX_RCV_SHIFT; 
	     maxloopcount >= 0;
	     maxloopcount--)
	{
		pFDBD = pDrvCtrl->FDAPtrCur;

		if (unlikely(MACOwnsRxFD(pFDBD)))
		{
			MSPEthReadIntSrcReg(pDrvCtrl);	/* Ensure data leave MAC to DXU */
			ReadMemConfig();	/* Ensure the data gets to SDRAM */
			if (MACOwnsRxFD(pFDBD))
			{
#ifdef DEBUGINTCOUNT
				rxpackets[pDrvCtrl->unit][count]++;
#endif			
				return 0;	/* If packet belongs to MAC */
			}
		}

#ifdef DEBUGINTCOUNT
		count++;
#endif			

		ReadMemConfig();	/* Ensure the data gets to SDRAM */
		status = pFDBD->FD.FDStat;

		/*
		 * A good or bad packet has been received.
		 */
		len = pFDBD->FD.FDLength;

		/* Include CRC for Ethernet in total */
		pDrvCtrl->stats.rx_bytes += len;
		pDrvCtrl->stats.rx_packets++;

		/* Remove CRC length from size of frame returned to user */
   		len -= ETHERNET_CRC_SIZE; 

#if 0
		{
			UCHAR		*p_buf;

			p_buf = (UCHAR *) pFDBD->BD[0].BuffData;
			PRINTK(KERN_INFO 
				"Packet received: %x, %d, %x: %x %x %x %x\n",
				(unsigned int) p_buf, 
				(unsigned int) len, 
				(unsigned int) status,
				*((unsigned int *)&p_buf[0]),
				*((unsigned int *)&p_buf[4]),
				*((unsigned int *)&p_buf[8]),
				*((unsigned int *)&p_buf[12]));
		}
#endif

		/* 
		 * Get a new skb and attach to the corresponding BD in the 
		 * BL area so that buffer is readily available for incoming packet. 
		 * And also the this function assumes only one BD is used per 
		 * incoming packet.
		 */
		bdNum = pFDBD->BD[0].BDStat;

		if (likely(status & RCVD_OK_PKT))	/* If it is a good packet */
		{
#ifdef	XCONFIG_NET_HW_FLOWCONTROL
			if (unlikely(netdev_dropping))
				goto throttle;
#endif	/* CONFIG_NET_HW_FLOWCONTROL */
			skb = pDrvCtrl->rskbp[bdNum];

			/* Assume sending packet up stack, start SDRAM fetch */
#if 1
			invalidate_dcache_line((u32)skb->data);
			pref(skb->data, 0x000);
			InvalidateBuffer(skb->data + 16, 256 - 16);
			pref(skb->data, 0x010);
//#ifndef	CONFIG_NET_FASTROUTE
//			pref(skb->data, 0x020);
//#endif	/* CONFIG_NET_FASTROUTE */
#else
			InvalidateBuffer(skb->data, len);
			pref(skb->data, 0x000);
			pref(skb->data, 0x010);
			pref(skb->data, 0x020);
#endif	/* 1 */

#ifdef	CONFIG_NET_FASTROUTE
			switch (msp_fast_route(pDrvCtrl->netdevice, skb,
				len, bdNum)) {
			case 0:
				goto gone;
			case 1:
				goto semi_gone;
			case 2:
				break;
			case 3:
				goto memory_squeeze;
			}
#endif	/* CONFIG_NET_FASTROUTE */

			pDrvCtrl->rskbp[bdNum] = getskb(pDrvCtrl);
			pref(skb->data, 0x020);

#ifdef	CONFIG_NET_FASTROUTE
memory_squeeze:
#endif	/* CONFIG_NET_FASTROUTE */

			if (unlikely(len > 256))
				InvalidateBuffer(skb->data + 256, len - 256);
			if (likely(pDrvCtrl->rskbp[bdNum] != NULL))
			{
				/* Finish skb and send it up */

				__skb_put(skb, len);

				skb->protocol = eth_type_trans(skb,
					pDrvCtrl->netdevice);
				pref(skb->data, 0x030);
#ifdef	CONFIG_NET_FASTROUTE
semi_gone:
#endif	/* CONFIG_NET_FASTROUTE */
				netif_rx(skb);
#ifdef	XCONFIG_NET_HW_FLOWCONTROL
				if (unlikely(netdev_dropping)) {
throttle:
					if (pDrvCtrl->fc_bit) {
						/* Do something to HW here */
						;
					}
				}
#endif	/* CONFIG_NET_HW_FLOWCONTROL */

#ifdef	CONFIG_NET_FASTROUTE
gone:
#endif	/* CONFIG_NET_FASTROUTE */
				pDrvCtrl->netdevice->last_rx = jiffies;
			}
			else
			{
				PRINTK(KERN_INFO
					"MSPEthReceive-dev_alloc_skb failed\n");
				/* MIB2 Statistic */
				pDrvCtrl->stats.rx_dropped++;	// Do all count as a dropped pkt??
				pDrvCtrl->rskbp[bdNum] = skb;
			}
		} else {
			msp_process_rx_error(pDrvCtrl, status);
		}

		{
			/* 
			 * Return ownership of frame and buffer descriptors
			 * in both FDA and BL. In case of Jabbers, there could
			 * be several BDs, so make sure that all the BDs that
			 * MAC used are cleared here.
			 */
			pFDBD = (FDBD *) &(pDrvCtrl->FDAPtrCur->FD); 
			bdNum = pFDBD->FD.FDCtl & FDCTL_BD_COUNT_MASK;
			if (likely(bdNum == 1)) {
				BD	*bd;
				int	bdix;

				bdix = pFDBD->BD[0].BDStat;
				bd = &pDrvCtrl->BLFrmPtr->BD[bdix];
				bd->BuffData =
					getRxBuffData(pDrvCtrl->rskbp[bdix]->data);
				bd->BuffLength = MSP_END_BUFSIZE;
				bd->BDCtl      = MAC_OWNS_BD;

				pFDBD->FD.FDCtl = MAC_OWNS_FD;
			} else {
				msp_rx_mult_bd(pDrvCtrl);
			}
		}

		/* Return free descriptor area back to controller */
		pDrvCtrl->FDAPtrCur = (FDBD *) pDrvCtrl->FDAPtrCur->FD.FDNext;
	}

#ifdef DEBUGINTCOUNT
	rxpackets[pDrvCtrl->unit][count]++;
#endif			
	return 1;		/* More RX descriptors to process */
}


/*********************************************************************
 * RXbh is called when the interrupts schedule it.
 */

static void __hotnet	RXbh(void *data)
{
	struct net_device	*dev = data;
	MSP_DEVICE	*pDrvCtrl = dev->priv;
	unsigned long	recovery;

	if (pDrvCtrl == 0)	/* If shutting down, do nothing */
		return;

	recovery = xchg_u32(&pDrvCtrl->bh_recovery, 0);

	if (pDrvCtrl->reloadflag)
		return;

	MSPEthRxIntEnable(pDrvCtrl);
	MSPEthReceive(pDrvCtrl);

	if (recovery)
	{
		MSPEthFdaFullIntEnable(pDrvCtrl);
		MSPEthBlFullIntEnable(pDrvCtrl);
	}

	if (! MACOwnsRxFD(pDrvCtrl->FDAPtrCur)) 
	{
		queue_task(pDrvCtrl->rx_bh, &tq_immediate);
		mark_bh(IMMEDIATE_BH);
	}
}


/************************************************************************
 * Open/initialize the board. This is called (in the current kernel)
 * sometime after booting when the 'ifconfig' program is run.
 *
 * This routine should set everything up anew at each open, even
 * registers that "should" only need to be set once at boot, so that
 * there is non-reboot way to recover if something goes wrong.
 */

static int	brecis_eth_open(struct net_device *dev)
{
	unsigned int	basePort = dev->base_addr;

	PRINTK(KERN_INFO "%s: Entering brecis_eth_open(), MSP_DEVICE=%x\n", 
		 dev->name, (unsigned int) dev->priv);

	/*
	 * This is used if the interrupt line can turned off (shared).
	 * - Can we share? - might as well register
	 */
	if (dev->priv == NULL)
	{
		/* If not already initialized */
		if (request_irq(dev->irq, &brecis_eth_interrupt, SA_SAMPLE_RANDOM, 
			    cardname, dev))
		{
			printk(KERN_INFO "%s: brecis_eth_open(), failed request_irq\n",
				dev->name);
			return -EAGAIN;
		}

		/* Grab the region so that no one else tries to probe our ioports. */
		if (!request_mem_region(basePort, MSPEthExtent, cardname))
		{
			printk(KERN_INFO
		  		"%s: brecis_eth_open(), failed request_mem_region\n", 
				dev->name);
		  	return -EAGAIN;
	  	}

		/* Already initialized in the Probe when allocated, do it again
		 * Initialize vars in private area- Again!
		 */
		brecis_eth_InitVars(dev, dev->base_addr);
	}

	/* Reset the hardware here. Don't forget to set the station address. */
	  
	if (MSP_SUCCESS != MSPEthSetup(dev))
	{
		printk(KERN_ERR "%s: brecismspeth: Phys init failed\n", dev->name);
		return -ENOMEM;
	}

	monitor_phy_negotiation();

	set_multicast_list(dev);

#ifdef	XCONFIG_NET_HW_FLOWCONTROL
	{
		MSP_DEVICE	*priv = dev->priv;

		priv->fc_bit = netdev_register_fc(dev, msp_xon);
	}
#endif	/* CONFIG_NET_HW_FLOWCONTROL */
#ifdef	CONFIG_NET_FASTROUTE
//	dev->tx_semaphore = 1;
#endif	/* CONFIG_NET_FASTROUTE */

	return 0;
}


static void	brecis_eth_complete_open(MSP_DEVICE *priv)
{
	memset(&priv->stats, 0, sizeof(struct net_device_stats));

	MSPEthAckAllInt(priv);
	MSPEthIntEnable(priv);
	MSPEthEnableTxRx(priv);

	/* We are now ready to accept transmit requests from
	 * the queueing layer of the networking.
	 */
	netif_start_queue(priv->netdevice);
}


/************************************************************************
 * brecis_eth_send_packet - Send a packet.
 *
 * This will only be invoked if your driver is _not_ in XOFF state.
 * What this means is that you need not check it, and that this
 * invariant will hold if you make sure that the netif_*_queue()
 * calls are done at the proper times.
 */

static int __hotnet	brecis_eth_send_packet(struct sk_buff *skb, struct net_device *dev)
{
	MSP_DEVICE	*priv = dev->priv;
	unsigned long	flags;
	FDBD		*FrmPtr;
  
	if (unlikely(priv == NULL || priv->reloadflag))
	{
		/* 
		 * It should be impossible to get here because whoever set
		 * priv to NULL or set the reloadflag should have done
		 * netif_stop_queue(), but stop us again
		 */
		printk(KERN_INFO
			"brecis_eth_send_packet -- priv == NULL or reloading\n");
		netif_stop_queue(dev);
		return -1;
	}

	/* If some error occurs while trying to transmit this
	 * packet, you should return '1' from this function.
	 * In such a case you _may not_ do anything to the
	 * SKB, it is still owned by the network queueing
	 * layer when an error is returned.  This means you
	 * may not modify any SKB fields, you may not free
	 * the SKB, etc.
	 */

	/* Finish and free any completed transmits if not using INTR_Tx */

	MSPEthFreeTxBuffs(priv);
  
	/* This is the most common case for modern hardware.
	 * The spinlock protects this code from the TX complete
	 * hardware interrupt handler.  Queue flow control is
	 * thus managed under this lock as well.
	 */
	spin_lock_irqsave(&priv->lock, flags);

	FrmPtr = priv->TxFrmPtrCur;

	if (unlikely(TxFDInUse(FrmPtr)))
	{
		netif_stop_queue(dev);
		MSPEthTxIntEnable(priv);
		spin_unlock_irqrestore(&priv->lock, flags);
		return 1;			/* We are busy */
	}

	priv->TxFrmPtrCur = (FDBD *)FrmPtr->FD.FDNext;

	FrmPtr->BD[0].BuffData = skb->data;
	FrmPtr->BD[0].BuffLength = skb->len;
	FrmPtr->FD.FDSystem = (u32) skb;

	FrmPtr->BD[0].BDCtl = MAC_OWNS_BD;
	FrmPtr->FD.FDCtl |= MAC_OWNS_FD;

	spin_unlock_irqrestore(&priv->lock, flags);

	/* Let the MAC know there is something to send */
	MSPEthTxWakeUp(priv);

	dev->trans_start = jiffies;

	return 0;
}


/************************************************************************
 * The typical workload of the driver:
 * Handle the network interface interrupts.
 */
static void	brecis_eth_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct net_device *dev = dev_id;
	MSP_DEVICE	*pDrvCtrl = (MSP_DEVICE *)dev->priv;
	int doreceive = 0;
	int dorecovery = 0;
	int status;

	if (unlikely(pDrvCtrl == 0)) {
		printk(KERN_INFO "%s: eth_int, pDrvCtrl=0!\n", dev->name);
		return;
	}

#ifdef DEBUGINTCOUNT
	if (pDrvCtrl->unit < MAX_DEVICES)
		intcount[pDrvCtrl->unit]++;
#endif

	status = MSPEthReadIntSrcReg(pDrvCtrl);

	ReadMemConfig();	/* Ensure the data gets to SDRAM */

	if (likely(status == INTR_Rx))
	{
		MSPEthAckRxInt(pDrvCtrl);
		MSPEthRxIntDisable(pDrvCtrl);
		if (! pDrvCtrl->reloadflag)
		{
			queue_task(pDrvCtrl->rx_bh, &tq_immediate);
			mark_bh(IMMEDIATE_BH);
		}
		return;
	}

#ifdef DEBUGJIFFIEINT
	if (debuginterruptcount >= MAXINTSPERJIFFIE)
	{
		/* 
		 * Disable ints, stop traffic, 
		 * arbitrarily ack all interrupts
		 */
		MSPEthMACIntDisable(pDrvCtrl);
		MSPEthDisableTxRxHalt(pDrvCtrl);
		MSPEthAckOtherInt(pDrvCtrl, status);

		if (debuginterruptcount == MAXINTSPERJIFFIE)
		{
			debuginterruptcount++;
			printk(KERN_INFO "at max ints per jiffie\n");
		}
		return;		/* be quiet */
	}

	if (debugjiffiecount != jiffies)
	{
		debugjiffiecount = jiffies;
		debuginterruptcount = 0;
	}
	else
	    debuginterruptcount++;

	debugintdev[debuginterruptcount] = pDrvCtrl->unit;
	debugintstat[debuginterruptcount] = status;

#endif
	if (unlikely(status & INTR_REQ_RELOAD))
	{
		MSPEthAckOtherInt(pDrvCtrl, status);

		PRINTK(KERN_ERR "brecis_eth_interrupt: dev<%x>->name<%s>, reload\n",
			   (unsigned int) dev, dev->name);

		setup_delayed_reload(dev);
		return;
	}

	if (likely(status & INTR_Tx))
	{
		MSPEthAckTxInt(pDrvCtrl);
		MSPEthFreeTxBuffs(pDrvCtrl);
	}
	
	if (unlikely(status & (INTR_FDAEx | INTR_IntFDAEx)))
	{
#ifdef DEBUG_RX_TX_STATS
		fdaExhausedCounts[pDrvCtrl->unit]++;
#endif
		doreceive = 1;
		dorecovery = 1;

		MSPEthFdaFullIntDisable(pDrvCtrl);

		if (status & INTR_FDAEx)
		{
			MSPEthAckFdaFullInt(pDrvCtrl);
		}
	}

	if (unlikely(status & (INTR_BLEx | INTR_IntBLEx)))
	{
		doreceive = 1;
		dorecovery = 1;

#ifdef DEBUG_RX_TX_STATS
		blExhausedCounts[pDrvCtrl->unit]++;
#endif
		MSPEthBlFullIntDisable(pDrvCtrl);

		if (status & INTR_BLEx)
		{
		    MSPEthAckBlFullInt(pDrvCtrl);
		}
	}

	if (likely(status & INTR_Rx))
	{
		doreceive = 1;

		/* Ack interrupt */
		
		MSPEthAckRxInt(pDrvCtrl);
	}
		
	if (likely(doreceive || dorecovery))
	{
		MSPEthRxIntDisable(pDrvCtrl);

		if (dorecovery)
			pDrvCtrl->bh_recovery = 1;

		if (!pDrvCtrl->reloadflag)
		{
			queue_task(pDrvCtrl->rx_bh, &tq_immediate);
			mark_bh(IMMEDIATE_BH);
		}
	}

	if (unlikely(status & INTR_OTHER))
	{
		MSPEthAckOtherInt(pDrvCtrl, status & INTR_OTHER);
	}
}


/************************************************************************
 * The inverse routine to net_open().
 */

static int	brecis_eth_close(struct net_device *dev)
{
	MSP_DEVICE	*priv = dev->priv;
	unsigned long	flags;

	PRINTK(KERN_INFO "%s: ***** Entering brecis_eth_close(), MSP_DEVICE=%x\n", 
		dev->name, (unsigned int) dev->priv);

	save_and_cli(flags);

#ifdef	XCONFIG_NET_HW_FLOWCONTROL
	if (priv != NULL && priv->fc_bit) {
		int	bit = priv->fc_bit;

		priv->fc_bit = 0;
		netdev_unregister_fc(bit);
	}
#endif	/* CONFIG_NET_HW_FLOWCONTROL */
	netif_carrier_off(dev);
	netif_stop_queue(dev);

	/* Flush the Tx and disable Rx here. */

	if (priv != NULL)
	{
		MSPEthMACIntDisable(priv);
		ReadMemConfig();
		MSPEthEnableTxHalt(priv);
		MSPEthEnableRxHalt(priv);
		NanoDelay(100000);	/* 100 uS */
		OrMacReg(priv, MAC_CTL, MAC_CTL_HaltReq);
		NanoDelay(30000);	/* 30 uS */
		MSPEthReadIntSrcReg(priv);
		MSPEthAckAllInt(priv);
		ReadMemConfig();
		WriteMacRegister(priv, MAC_CTL, MAC_CTL_Reset);
		NanoDelay(10000);	/* 10 uS */
		MSPEthReset(priv);
		MSPEthMemoryFree(priv);
	}

	free_irq(dev->irq, dev);

	restore_flags(flags);

	release_mem_region(dev->base_addr, MSPEthExtent);

#if	0	/*def	CONFIG_NET_FASTROUTE*/
//	dev->tx_semaphore = 0;
	if (priv != NULL) {
		struct sk_buff	*skb = xchg(&priv->skb_cache, NULL);
		if (skb)
			kfree_skb(skb);
	}
#endif	/* CONFIG_NET_FASTROUTE */

#ifdef DEBUGJIFFIEINT
	{
		int i;

		printk(KERN_INFO "%s: debug interruptcount %d, jiffie count %d\n",
			dev->name, debuginterruptcount, debugjiffiecount);

		for (i = 0; i < debuginterruptcount; i++)
			printk(KERN_INFO "%s: %08d) %08x  %08x\n",
				dev->name, i, debugintdev[i], debugintstat[i]);
	}
#endif  

	return 0;
}


/**************************************************************************
 * Reload ethernet MAC interface because we have a problem or lost carrier.
 */

static void	reload_tq(void *data)
{
	struct net_device	*dev = data;

	if (dev->priv == 0)	/* If already shutting down, do nothing */
		return;

	brecis_eth_close(dev);
	brecis_eth_open(dev);
	netif_wake_queue(dev);
}


static void reloadcheck_timer(unsigned long unused)
{
	tasklet_schedule(&mspeth_reloadcheck_task);
	mspeth_reloadcheck_timer.expires = jiffies + RELOAD_CHECK_DELAY * HZ;
	add_timer(&mspeth_reloadcheck_timer);
}


static void	reloadchecks(unsigned long unused)
{
	int		i;
	MSP_DEVICE	*priv;
	struct net_device	*dev;

	/*
	 * For every ethernet interface, check for carrier
	 * and signal a change link event if necessary
	 */
	for (i = 0; i < MSP_ETH_MAX_UNITS; i++)
	{
		priv = gEthDrvCtrl[i];

		if (priv != NULL)
		{
			dev = priv->netdevice;

			/* Try to free completed tx buffs */

			if (priv->reloadflag)
			{
				/* If reload is being forced */

				schedule_task(priv->reload_tq);
				continue;
			}

			if (jiffies - dev->last_rx < dev->watchdog_timeo)
			{
				/*
				 * We either got a packet (last_rx)
				 * within the last watchdog_timeo interval 
				 * so this interface was alive so don't 
				 * have to check cable this time
				 */
				continue;
			}

			MSPEthFreeTxBuffs(priv);

			if (TxFDInUse(priv->TxFrmPtrFreeBuf) &&
			    jiffies - dev->trans_start > dev->watchdog_timeo &&
			    priv->tx_packets == priv->stats.tx_packets &&
			    priv->tx_errors == priv->stats.tx_errors)
			{
				/* 
				 * If we have a packet in use -and-
				 * the jiffies check makes sure no TX packet
				 * was just put on the interface -and-
				 * we haven't finished any tx packet,
				 * then we timed out.
				 */
				printk(KERN_INFO 
				       "%s: tx packet timed out\n",
				       dev->name);
				schedule_task(priv->reload_tq);
				continue;
			}

			priv->tx_packets = priv->stats.tx_packets;
			priv->tx_errors = priv->stats.tx_errors;

			if (priv->phyType == PHYTYPE_KENDINSWITCH
				||  priv->phyType == PHYTYPE_NONE)
			{
				/* A switch always HAS carrier */
				continue;
			}

			if (priv->phyWaiting)
			{
				/* Found a PHY coming up */
				continue;
			}

			if (MSPPHYGCheckCable(priv) != 0)
			{
				/* We have a carrier */

				if (! netif_carrier_ok(dev))
				{
					/* If need to notify upper layer */

					if ((jiffies - priv->carrier_off) >
					    CARRIER_RELOAD_DELAY)
					{
						/* Carrier was off too long */

						schedule_task(priv->reload_tq);
					}
					else
					{
						/* Only tell upper layer */
						netif_carrier_on(dev);
					}
				}
			}
			else
			{
				/* We don't have a carrier */

				if (netif_carrier_ok(dev))
				{
					/* If need to notify upper layer */

					netif_carrier_off(dev);
					priv->carrier_off = jiffies;
				}
			}
		}
	}
}


/**************************************************************************
 * Monitor PHY Negotation.
 */

static void	monitor_phy_negotiation(void)
{
	if (timer_pending(&mspeth_checkphys_timer))
		return;

	tasklet_schedule(&mspeth_checkphys_task);
}


static void	checkphys_timer(unsigned long unused)
{
	tasklet_schedule(&mspeth_checkphys_task);
}


static void	checkphys(unsigned long unused)
{
	int		i;
	int		needphycheck = 0;
	MSP_DEVICE	*priv;

	/*
	 * For every ethernet interface, check for PHY negotation
	 */
	for (i = 0; i < MSP_ETH_MAX_UNITS; i++)
	{
		priv = gEthDrvCtrl[i];
		
		if (priv != NULL)
		{
			if (priv->phyWaiting)
			{
				/* found a PHY coming up */

				needphycheck = 1;

				switch (MSPPHYGWaitForLink(priv))
				{
				case MSP_SUCCESS:
					MSPEthPhyInit_complete(priv);
					brecis_eth_complete_open(priv);
					break;
				case MSP_FAIL:
					break;
				}
			}
		}
	}

	if (needphycheck)
	{
		mspeth_checkphys_timer.expires = jiffies + 1;
		add_timer(&mspeth_checkphys_timer);
	}
}


/************************************************************************
 * Get the current statistics.
 * This may be called with the card open or closed.
 */

static struct net_device_stats	*brecis_eth_get_stats(struct net_device *dev)
{
	MSP_DEVICE	*priv = (MSP_DEVICE *)dev->priv;

	if (priv == NULL)
		return NULL;

#ifdef DEBUGDISPLAYMACREGISTERS
	MSPEthDisplayMacRegisters(priv->unit);
#endif
#ifdef DEBUG_RX_TX_STATS
	printk(KERN_INFO "Packet stats - unit %d\n", (UINT) priv->unit);
	printk(KERN_INFO "rx_len_errors    %d\n", (UINT)priv->stats.rx_length_errors);
	printk(KERN_INFO "rx_crc_errors    %d\n", (UINT)priv->stats.rx_crc_errors);
	printk(KERN_INFO "rx_over_errors   %d\n", (UINT)priv->stats.rx_over_errors);
	printk(KERN_INFO "rx_fifo_errors   %d\n", (UINT)priv->stats.rx_fifo_errors);
	printk(KERN_INFO "rx_frame_errors  %d\n", (UINT)priv->stats.rx_frame_errors);
	printk(KERN_INFO "rx_missed_errors %d\n", (UINT)priv->stats.rx_missed_errors);
	printk(KERN_INFO "fdaExhausedCounts %d\n", (UINT)fdaExhausedCounts[priv->unit]);
	printk(KERN_INFO "blExhausedCounts %d\n", (UINT)blExhausedCounts[priv->unit]);
					
	printk(KERN_INFO "tx_collisions       %d\n", (UINT)priv->stats.collisions);				
	printk(KERN_INFO "tx_aborted_errors   %d\n", (UINT)priv->stats.tx_aborted_errors);				
	printk(KERN_INFO "tx_carrier_errors   %d\n", (UINT)priv->stats.tx_carrier_errors);				
	printk(KERN_INFO "tx_fifo_errors      %d\n", (UINT)priv->stats.tx_fifo_errors);				
	printk(KERN_INFO "tx_heartbeat_errors %d\n", (UINT)priv->stats.tx_heartbeat_errors);				
#endif

#ifdef DEBUGINTCOUNT
	{
		int i;

		for (i = 0; i < DEBUGINTCOUNT; i++)
			if (rxpackets[priv->unit][i])
				printk(KERN_INFO 
					"eth%d: rx num %d time %d\n", 
					priv->unit, i, rxpackets[priv->unit][i]);
		printk(KERN_INFO "eth%d: %d ints\n", 
			priv->unit, intcount[priv->unit]);
	}
#endif

	return &priv->stats;
}


/************************************************************************
 * Set hardware MAC address
 */
static int	brecis_eth_set_mac_address(struct net_device *dev, void *addr)
{
	struct sockaddr *mac = addr;

	memcpy(dev->dev_addr, mac->sa_data, EADDR_LEN);
	set_multicast_list(dev);
	
	return 0;
}


#if 1
/************************************************************************
 * Perform ioctls.
 */
static int brecis_eth_do_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	int	result = 0;
        MSP_DEVICE      *pDrvCtrl;
        u32 *data = (u32 *) & rq->ifr_data;
        pDrvCtrl = dev->priv;

	switch (cmd)
	{
	case SIOCDEVPRIVATE + 9:
		if (MSPPHYGCheckCable(pDrvCtrl) != 0) {
			printk("brecis_eth_do_ioctl(): cable on\n");
			*data =  1;
		} else {
			printk("brecis_eth_do_ioctl(): cable off\n");
			*data =  0;
			/* by jkglee, brecis's kernel function copytouser not work? 
			 * check return code instead, while ioctl 
			 * return 0         : connector on
			 * return -19XX     : connector off */
			result = -1923;
		}
		break;
	default:
		result = -EOPNOTSUPP;
		break;
	}

	return result;
}
#endif


/************************************************************************
 * Set or clear the multicast filter for this adapter.
 * num_addrs == -1	Promiscuous mode, receive all packets
 * num_addrs == 0	Normal mode, clear multicast list
 * num_addrs > 0	Multicast mode, receive normal and MC packets,
 *			and do best-effort filtering.
 */

#define	ARC_MC1		0x80	/* Offset to MC1 entry */
#define	ARC_MC2		0x84	/* Offset to MC2 entry */
#define	MAXMCADDR	18	/* Maximum number of multicast entries */

static void
set_multicast_list(struct net_device *dev)
{
	int	swEntryNumber;
	MSP_DEVICE	*priv = (MSP_DEVICE *)dev->priv;
	struct dev_mc_list	*ptr = dev->mc_list;
	int	i;
	int	all_multi = 0;
//	union { u8 b[ARC_SIZE]; u32 w[ARC_SIZE / sizeof(u32)]; } ARCData;

	PRINTK(KERN_INFO "***** Entering set_multicast_list, %s, MSP_DEVICE=%x\n", 
		dev->name, (unsigned int) dev->priv);

	memcpy(&priv->ARCData.b[0*6], "\x01\x80\xC2\x00\x00\x01", EADDR_LEN);
	memcpy(&priv->ARCData.b[1*6], dev->dev_addr, EADDR_LEN);
	memcpy(&priv->ARCData.b[20*6], "\x88\x08\x00\x01\x00\x10\x00\x00", 8);

	if (priv == NULL) {
		printk("NULL\n");
		return;
	}

	if (dev->flags & IFF_PROMISC)
	{
		PRINTK(KERN_INFO "set_multicast_list IFF_PROMISC\n");
		MSPEthEnablePromiscMode(priv);
		return;
	}

	WriteMacRegister(priv, ARC_ENA, 0);
	WriteMacRegister(priv, ARC_CTL, 0);

	swEntryNumber = 0;

	if (dev->flags & IFF_ALLMULTI)
	{
		PRINTK(KERN_INFO "set_multicast_list IFF_ALLMULTI\n");
		WriteMacRegister(priv, ARC_CTL, 
			ARC_CompEn | ARC_BroadAcc | ARC_GroupAcc);
		return;
	}

	for (ptr = dev->mc_list; ptr != NULL; ptr = ptr->next)
	{
		if (swEntryNumber >= MAXMCADDR) {
			all_multi = 1;
			break;
		}
		memcpy(&priv->ARCData.b[6 * (2 + swEntryNumber)], ptr->dmi_addr,
			EADDR_LEN);
		++swEntryNumber;
	}

	for (i = 0; i < ARC_SIZE; i += 4) {
		WriteMacRegister(priv, ARC_ADR, i);
		WriteMacRegister(priv, ARC_DATA, priv->ARCData.w[i >> 2]);
	}

	WriteMacRegister(priv, ARC_ADR, 0);
	WriteMacRegister(priv, ARC_ENA,
		(((1 << swEntryNumber) - 1) << 2) | (1 << 1) | (1 << 0));

	if (all_multi) {
		PRINTK(KERN_INFO "eth%d: set_multicast_list MAXMCADDDR\n");
		WriteMacRegister(priv, ARC_CTL,
			ARC_CompEn | ARC_BroadAcc | ARC_GroupAcc);
	} else {
		PRINTK(KERN_INFO "set_multicast_list normal reception\n");
		WriteMacRegister(priv, ARC_CTL, ARC_CompEn | ARC_BroadAcc);
	}

	PRINTK(KERN_INFO "**** Exiting set_multicast_list, %s, MSP_DEVICE=%p\n",
		dev->name, dev->priv);
}



/************************************************************************
 * Set up initialization/cleanup
 */

static void __init	init_phyaddresses(void)
{
	int	i;
	int	hwindex;
	int	phyindex;
	char	*phyaddr;
	char	name[80];

	for (i = 0; i < MSP_ETH_MAX_UNITS; i++)
	{
		sprintf(name, "phyaddr%d", i);
		phyaddr = prom_getenv(name);

		if (phyaddr != NULL &&
		    sscanf(phyaddr, "%d:%d", &hwindex, &phyindex) == 2 &&
		    hwindex < MSP_ETH_MAX_UNITS && 
		    (phyindex < MD_MAX_PHY || phyindex == MD_DYNAMIC_PHY) &&
		    PhyBases[hwindex].mac_base != NULL)
		{
			phyAddresses[i] = (hwindex << 16) | phyindex;
		}
		else if (gethswitch[i])
		{
			phyAddresses[i] = MD_NO_PHY;
		}
		else
		{
			/* If Polo, assume we use mdc/mdio index 0 hardware */
			/* Otherwise, assume each phy has own mdc/mdio */

			switch (identify_family())
			{
			case FAMILY_POLO:
			case FAMILY_ZEUS:
				phyAddresses[i] = MD_DYNAMIC_PHY;
				break;
			
			default:
				phyAddresses[i] = (i << 16) | MD_DYNAMIC_PHY;
			}

			if (phyaddr != NULL)
			{
				pkinit(KERN_INFO 
				       "init_phyaddresses: %s bad value %s\n",
				       name, phyaddr);
			}
		}
	}
}

/* by jkglee */
#if 0
static void __init	init_speedduplexswitchsettings(void)
{
	int	index;
	int	unit;
	char	c = ' ';
	char	command_line[COMMAND_LINE_SIZE];
	char	*ptr = &command_line[0];
	char	*ethptr = NULL;

	strcpy(command_line, prom_getcmdline());

	while (c != '\0')
	{
		if (c != ' ' || memcmp(ptr, "ip=", 3) != 0)
		{
			c = *ptr++;
			continue;
		}

		c = *ptr++;
		index = 0;
		unit = -1;

		while (index < 8)
		{
			c = *ptr++;

			if (c == '\0' || c == ' ')
			{
				if (index == 7)
				{
					index++;
					*--ptr = '\0';
					ptr++;
				}

				break;
			}

			if (c == ':')
			{
				index++;

				if (index == 5)
				{
					if (memcmp(ptr, "eth", 3) != 0)
					{
						break;
					}
					
					ethptr = &ptr[3];
					ptr = ethptr;
				}

				if (index == 6)
				{
					*--ptr = '\0';
					ptr++;
					unit = simple_strtol(ethptr, NULL, 0);
				}

				if (index == 7)
				{
					ethptr = ptr;
				}

				if (index == 8)
				{
					*--ptr = '\0';
					ptr++;
				}
			}
		}

		if (index < 8 || unit < 0 || unit > MSP_ETH_MAX_UNITS)
			continue;

		if (strcmp(ethptr, "an") == 0)
			gethspeedduplex[unit] = -1;
		else if (strcmp(ethptr, "10h") == 0)
			gethspeedduplex[unit] = 10;
		else if (strcmp(ethptr, "10f") == 0)
			gethspeedduplex[unit] = 20;
		else if (strcmp(ethptr, "100h") == 0)
			gethspeedduplex[unit] = 100;
		else if (strcmp(ethptr, "100f") == 0)
			gethspeedduplex[unit] = 200;
		else if (strcmp(ethptr, "10hs") == 0)
		{
			gethspeedduplex[unit] = 10;
			gethswitch[unit] = 1;
		}
		else if (strcmp(ethptr, "10fs") == 0)
		{
			gethspeedduplex[unit] = 20;
			gethswitch[unit] = 1;
		}
		else if (strcmp(ethptr, "100hs") == 0)
		{
			gethspeedduplex[unit] = 100;
			gethswitch[unit] = 1;
		}
		else if (strcmp(ethptr, "100fs") == 0)
		{
			gethspeedduplex[unit] = 200;
			gethswitch[unit] = 1;
		}
	}
}
#endif


static int __init mspeth_init(void)
{
	init_timer(&mspeth_reloadcheck_timer);
	mspeth_reloadcheck_timer.expires = jiffies + RELOAD_CHECK_DELAY * HZ;
	mspeth_reloadcheck_timer.function = reloadcheck_timer;
	mspeth_reloadcheck_timer.data = 0;
	add_timer(&mspeth_reloadcheck_timer);

	init_timer(&mspeth_checkphys_timer);
	mspeth_checkphys_timer.expires = 0;
	mspeth_checkphys_timer.function = checkphys_timer;
	mspeth_checkphys_timer.data = 0;
	switch (identify_family())
	{
	case FAMILY_POLO:
	case FAMILY_ZEUS:
		brctl_present = 1;
		break;
	
	default:
		break;
	}

#if 0
	init_speedduplexswitchsettings();
#endif
	/* by jkglee, 2003.10.24. eth0 auto negotiate with phy */
        gethspeedduplex[0] = 0;
        gethswitch[0] = 0;

	/* eth1 100 full with switch */
        gethspeedduplex[1] = 200;
        gethswitch[1] = 1;
	
	init_phyaddresses();
	mspeth_inited = 1;
	return 0;
}

module_init(mspeth_init);

/*
 * Local variables:
 *  version-control: t
 *  kept-new-versions: 5
 *  tab-width: 8
 *  c-indent-level: 0
 * End:
 */
#endif // CONFIG_BRECIS_ETH
