/*
 $Header: /home/cvsroot/B49G/linux-brecis-2.4.20/drivers/net/phyethgeneric.c,v 1.1.1.1 2004/04/29 09:53:56 jkglee Exp $
 */

/******************************************************************
 * Copyright (c) 2002,2003 BRECIS Communications
 *
 *     This program is free software; you can redistribute it
 *     and/or modify it under the terms of the GNU General
 *     Public License as published by the Free Software
 *     Foundation; either version 2 of the License, or (at your
 *     option) any later version.
 *
 *     This program is distributed in the hope that it will be
 *     useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
 *     02139, USA.
 *
 * BRECIS COMMUNICATIONS DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS
 * SOFTWARE.
 */

#include <asm/brecis/prom.h>
#include "brecismspeth.h"
#include "phyethgeneric.h"

static int	selectPhyRegister(MSP_DEVICE *pDrvCtrl, int reg)
{
	int	i;

	WrPhy(pDrvCtrl, MD_CA, (pDrvCtrl->macPhyId | MD_CA_BUSY_BIT | reg));

	i = 0;
	do
	{
		udelay(MD_REGISTER_UDELAY);
		if (((RdPhy(pDrvCtrl, MD_CA)) & MD_CA_BUSY_BIT) == 0)
			return 0;

		i += MD_REGISTER_UDELAY;
	} while (i < PHYG_UDELAY_FOR_RESPONSE);

	return 1;
}


int	readPhyRegister(MSP_DEVICE *pDrvCtrl, int reg)
{
	unsigned long	flags;
	int	data;

	spin_lock_irqsave(&pDrvCtrl->phy_base->lock, flags);

	if (selectPhyRegister(pDrvCtrl, reg))
	{
		spin_unlock_irqrestore(&pDrvCtrl->phy_base->lock, flags);
		if (! pDrvCtrl->phyprobe)
			printk(KERN_INFO "eth%d: phy %d, reg %d no response\n",
				pDrvCtrl->unit, pDrvCtrl->macPhyId >> 5, reg);
		return 0;
	}

	data = RdPhy(pDrvCtrl, MD_DATA);

	spin_unlock_irqrestore(&pDrvCtrl->phy_base->lock, flags);

	return data;
}


int	writePhyRegister(MSP_DEVICE *pDrvCtrl, int reg, int value)
{
	unsigned long	flags;

	spin_lock_irqsave(&pDrvCtrl->phy_base->lock, flags);

	WrPhy(pDrvCtrl, MD_DATA, value);

	if (selectPhyRegister(pDrvCtrl, reg | MD_CA_Wr))
	{
		spin_unlock_irqrestore(&pDrvCtrl->phy_base->lock, flags);
		return 1;
	}

	spin_unlock_irqrestore(&pDrvCtrl->phy_base->lock, flags);

	return 0;
}


/*************************************************************************
 * MSPPHYGCheckCable - Check for Cable connection to the PHY 
 *
 * RETURNS: non-zero if cable is okay or no PHY or we can't read PHY register
 */

int	MSPPHYGCheckCable(MSP_DEVICE *pDrvCtrl)
{
	int	data;

	switch (pDrvCtrl->phyType)
	{
	case PHYTYPE_NONE:
	case PHYTYPE_KENDINSWITCH:
		return 1;	/* No PHY, cable is always okay */

	default:
		break;
	}

	data = readPhyRegister(pDrvCtrl, PHYG_STATUS_REG_ONE);

	return (data & PHYGb_STATUS_REG_ONE_LINK_STATUS) != 0;
}


/*************************************************************************
 * MSPPHYGGetSpeedAndDuplex - Gets the Working speed from the PHY 
 *
 * RETURNS: MSP_SUCCESS or ERROR 
 */

int	MSPPHYGGetSpeedAndDuplex(MSP_DEVICE *pDrvCtrl)
{
	int	reg4;
	int	reg5;
	int	data;

	/* Read Register 0 (PHYG_CONTROL) */
	data = readPhyRegister(pDrvCtrl, PHYG_CONTROL);
	if (data == 0)
		return MSP_MAC_PHY_ERROR;

	if ((data & PHYGb_CONTROL_AUTONEG) == 0)
	{
		printk(KERN_INFO "eth%d: PHY is MANUALLY CONFIGURED for ",
			pDrvCtrl->unit);
		if (data & PHYGb_CONTROL_SELECT_SPEED)
		{
			pDrvCtrl->speed = MSPEth_100MBS;
		} else {
			pDrvCtrl->speed = MSPEth_10MBS;
		}
		printk("%d Mbps, ", pDrvCtrl->speed / 1000000);

		if (data & PHYGb_CONTROL_DUPLEXMODE)
		{
			printk("Full Duplex\n");
			pDrvCtrl->duplexMode = MAC_DPX_MODE_FULL;
		}
		else
		{
			printk("Half Duplex\n");
			pDrvCtrl->duplexMode = MAC_DPX_MODE_HALF;
		}
		return MSP_SUCCESS;
	}

	printk(KERN_INFO "eth%d: PHY is in AUTO NEGOTIATION MODE\n",
		pDrvCtrl->unit);
	/* doing auto negotiation */
	/* Read Register 4 (PHYG_AUTO_NEG_ADVERTISING) */
	reg4 = readPhyRegister(pDrvCtrl, PHYG_AUTO_NEG_ADVERTISING);

	/* Read Register 5 (PHYG_AUTO_NEG_LINK_PARTNER_ABILITY) */
	reg5 = readPhyRegister(pDrvCtrl, PHYG_AUTO_NEG_LINK_PARTNER_ABILITY);

	data = reg4 & reg5;

	if (data & PHYGb_AUTO_NEG_ADVERT_100BASE_TX_FULL)
	{
		pDrvCtrl->speed = MSPEth_100MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_FULL;
	}
	else if (data & PHYGb_AUTO_NEG_ADVERT_100BASE_T4)
	{
		pDrvCtrl->speed = MSPEth_100MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_HALF;
	}
	else if (data & PHYGb_AUTO_NEG_ADVERT_100BASE_TX)
	{
		pDrvCtrl->speed = MSPEth_100MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_HALF;
	}
	else if (data & PHYGb_AUTO_NEG_ADVERT_10BASE_T_FULL)
	{
		pDrvCtrl->speed = MSPEth_10MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_FULL;
	}
	else /*if (data & PHYGb_AUTO_NEG_ADVERT_10BASE_T)*/
	{
		pDrvCtrl->speed = MSPEth_10MBS;
		pDrvCtrl->duplexMode = MAC_DPX_MODE_HALF;
	}

	printk(KERN_INFO "eth%d: PHY is set for %dMbps %s Duplex\n",
		pDrvCtrl->unit, pDrvCtrl->speed / 1000000,
		pDrvCtrl->duplexMode == MAC_DPX_MODE_FULL ?
		"Full" : "Half");

	return MSP_SUCCESS;
}


int	MSPPHYGWaitForLink(MSP_DEVICE *pDrvCtrl) 
{
	int	data;

	/* Wait for link or auto negotiation to complete */

	/* Read Register 1 (PHYG_STATUS_REG_ONE) */
	data = readPhyRegister(pDrvCtrl, PHYG_STATUS_REG_ONE);
	if (data == 0)
	{
		pDrvCtrl->phyWaiting = 0;
		return MSP_MAC_PHY_ERROR;
	}

	if (pDrvCtrl->phyAutoNegot)
	{
		if (data & PHYGb_STATUS_REG_ONE_AUTO_NEG_DONE) 
		{
			printk(KERN_INFO "eth%d: AUTO NEGOTIATION COMPLETE\n",
				pDrvCtrl->unit);
			pDrvCtrl->phyWaiting = 0;
			return MSPPHYGGetSpeedAndDuplex(pDrvCtrl);
		}
	}
	else if (data & PHYGb_STATUS_REG_ONE_LINK_STATUS)
	{
		printk(KERN_INFO "eth%d: LINK DETECTED\n", pDrvCtrl->unit);
		pDrvCtrl->phyWaiting = 0;
		return MSPPHYGGetSpeedAndDuplex(pDrvCtrl);
	}

	if (--pDrvCtrl->phyWaiting == 0)
	{
		printk(KERN_INFO "eth%d: NO LINK DETECTED\n",
		       pDrvCtrl->unit);
		return MSP_FAIL;
	}

	return MSP_MAC_PHY_NO_LINK;
}


/*************************************************************************
 * MSPPHYGSetup - initialize and configure the PHY device.
 * This routine initialize and configure the PHY device 
 *
 * RETURNS: OK, or ERROR
 */

void	MSPPHYGSetup(MSP_DEVICE *pDrvCtrl) 
{
	int	data;

	printk(KERN_INFO
		"eth%d: Using PHY address %x with base address %p\n",
		pDrvCtrl->unit, pDrvCtrl->macPhyId >> 5, 
		pDrvCtrl->phy_base->mac_base);

	/*
	 * Loopback=disabled
	 * Speed = 100/10
	 * Auto-Negotiation = Disabled
	 * Power Down = 0 normal
	 * Isolate = 0 normal
	 * Restart Auto-Negotiation = 0
	 * Duplex = 1 Full / 0 Half
	 * Collision Test = 0 Disabled
	 *
	 * For Full Dup 10Mbps:  0x0100
	 * For Half Dup 10Mbps:  0x0000
	 * For Full Dup 100Mbps: 0x2100
	 * For Half Dup 100Mbps: 0x2000
	 *
	 * Auto Negotiation Enabled: 0x1000
	 */
	printk(KERN_INFO "eth%d: PHY is set for ", pDrvCtrl->unit);
	switch (pDrvCtrl->ethMode)
	{
	default:
	case ETH_AUTO:
		printk("Auto-Negotiation\n");
		data = PHYGb_CONTROL_AUTONEG;
		break;

	case ETH_10Mbps_HALF:
		printk("10 Mbps Half Duplex\n");
		data = PHYGb_CONTROL_SPEED_10_HALF;
		break;

	case ETH_10Mbps_FULL:
		printk("10 Mbps Full Duplex\n");
		data = PHYGb_CONTROL_SPEED_10_FULL;
		break;

	case ETH_100Mbps_HALF:
		printk("100 Mbps Half Duplex\n");
		data = PHYGb_CONTROL_SPEED_100_HALF;
		break;

	case ETH_100Mbps_FULL:
		printk("100 Mbps Full Duplex\n");
		data = PHYGb_CONTROL_SPEED_100_FULL;
		break;	
	}

	/* Send the data to the PHY */
   	if (writePhyRegister(pDrvCtrl, PHYG_CONTROL, data))
	{
		printk(KERN_INFO "MSPPHYGSetup(): PHY BUSY Setting Speed/Mode\n");
		return;
	}

	/* Read Register 0 (PHYG_CONTROL) */
	data = readPhyRegister(pDrvCtrl, PHYG_CONTROL);
	if (data == 0)
		return;

	pDrvCtrl->phyAutoNegot = (data & PHYGb_CONTROL_AUTONEG) != 0;

	pDrvCtrl->phyWaiting = MAX_TIME_TO_WAIT_FOR_LINK;
}



void	MSPPHYGIsolateON(MSP_DEVICE *pDrvCtrl)
{
	int	data;

	data = readPhyRegister(pDrvCtrl, PHYG_CONTROL);

	data |= PHYGb_CONTROL_ISOLATE;

	/* Send the data to the PHY */
   	if (writePhyRegister(pDrvCtrl, PHYG_CONTROL, data))
	{
		printk(KERN_INFO "MSPEthIsolatePHYON(): PHY BUSY\n");
		return;
	}
}


void	MSPPHYGIsolateOFF(MSP_DEVICE *pDrvCtrl)
{
	int	data;

	data = readPhyRegister(pDrvCtrl, PHYG_CONTROL);

	data &= ~PHYGb_CONTROL_ISOLATE;

	/* Send the data to the PHY */
   	if (writePhyRegister(pDrvCtrl, PHYG_CONTROL, data))
	{
		printk(KERN_INFO "MSPPHYGIsolate: PHY BUSY\n");
		return;
	}
}

