/* $Header: /home/cvsroot/B49G/linux-brecis-2.4.20/include/brecis/msp_vpm.h,v 1.1.1.1 2004/04/29 09:54:07 jkglee Exp $ */
/*
** Definitions for Brecis voice engine.  User API portion.
**
** Copyright 2003 Brecis Communications
**
*/

#ifndef BRECIS_MSP_VE_H
#define BRECIS_MSP_VE_H

/*
 * The following structure is returned by the BRECIS_VPM_CODEC_STATUS
 * ioctl(). It closely reflects the data returned from the VPM
 * as a result of the DIVPM_PB_SLOT_STAT command, some of the
 * data is redundant (frame_size, fpp), some is redundant *and*
 * reported in Brecis-speak, as opposed to Linux-speak (rx_codec_type,
 * tx_codec_type), some is reported in obscure Brecis-speak
 * (frame_type.)
 */
typedef	struct {
	unsigned int	port_enable ;
	unsigned int	decode_enable ;
	unsigned int	rx_codec_type ;
	unsigned int	tx_codec_type ;
	unsigned int	frame_type ;
	unsigned int	vad_status ;
	unsigned int	cng_status ;
	unsigned int	frame_size ;
	unsigned int	fpp ;
	unsigned int	packets_inside_jitter_buf ;
	unsigned int	samples_inside_playout_buf ;
	unsigned int	one_ms_insert ;
	unsigned int	one_ms_discard ;
} MSP_VPM_CODEC_STAT ;

/*
 * DTMF string generation structure.
 * The maximum string length is constrained
 * by the VPM firmware.
 */
#define	MSP_VPM_MAX_DTMF_LEN		32
typedef struct {
	unsigned int	length ;
	unsigned char	string[MSP_VPM_MAX_DTMF_LEN] ;
} MSP_VPM_DTMF_STRING ;

/*
 * Structure used when joining/quitting conference
 * calls.
 *
 * During a join command, target specifies the channel we
 * wish to share audio with. It is ignored during a quit
 * command.
 *
 * The which field specifies whether the tdm port, the codec
 * port, or both will be added to/removed from the conference.
 */
typedef struct {
	unsigned int	target ;
	unsigned int	which ;
#define	MSP_VPM_CONF_TDM	0x01
#define	MSP_VPM_CONF_CODEC	0x02
} MSP_VPM_CONFERENCE ;

/*
 * Brecis voice ioctls use 'B' and start at 0x20
 * (note: security engine uses some number of
 * ioctls starting at 0x01.)
 */
#define BRECIS_VPM_SET_TDM_PORT		_IOR('B', 0x20, int)
#define	BRECIS_VPM_ECHO_TONE_DISABLE	_IOWR('B', 0x21, int)
#define	BRECIS_VPM_NLP			_IOWR('B', 0x22, int)
#define	BRECIS_VPM_CNG			_IOWR('B', 0x23, int)
#define	BRECIS_VPM_TDM_DTMF_DETECT	_IOWR('B', 0x24, int)
#define	BRECIS_VPM_VAD			_IOWR('B', 0x25, int)
#define	BRECIS_VPM_DTMF_RELAY		_IOWR('B', 0x26, int)
#define	BRECIS_VPM_CALL_CLASS		_IOWR('B', 0x27, int)
#define	BRECIS_VPM_DOUBLE_TALK_DETECT	_IOWR('B', 0x28, int)
#define	BRECIS_VPM_ECHO_CANCELLER	_IOWR('B', 0x29, int)
#define	BRECIS_VPM_RX_LATENCY		_IOWR('B', 0x2a, int)
#define	BRECIS_VPM_REORDER		_IOR('B', 0x2b, int)
#define	BRECIS_VPM_CID_START		_IOWR('B', 0x2c, int)
#define	BRECIS_VPM_CIDCW		_IOW('B', 0x2d, PHONE_CID *)
#define	BRECIS_VPM_CODEC_STATUS		_IOR('B', 0x2e, MSP_VPM_CODEC_STAT *)
#define	BRECIS_VPM_DTMF_PLAY		_IOW('B', 0x2f, MSP_VPM_DTMF_STRING *)
#define	BRECIS_VPM_RECALL		_IOR('B', 0x30, int)
#define	BRECIS_VPM_MSG_WAITING		_IOR('B', 0x31, int)
#define	BRECIS_VPM_CONFIRMATION		_IOR('B', 0x32, int)
#define	BRECIS_VPM_ROH			_IOR('B', 0x33, int)
#define	BRECIS_VPM_CALL_WAIT_1		_IOR('B', 0x34, int)
#define	BRECIS_VPM_CALL_WAIT_2		_IOR('B', 0x35, int)
#define	BRECIS_VPM_CALL_WAIT_3		_IOR('B', 0x36, int)
#define	BRECIS_VPM_CALL_WAIT_4		_IOR('B', 0x37, int)
#define	BRECIS_VPM_CALIBRATION		_IOR('B', 0x38, int)
#define	BRECIS_VPM_DTMF_LONG_READY	_IOR('B', 0x39, int)
#define	BRECIS_VPM_GET_DTMF_LONG_ASCII	_IOR('B', 0x3a, int)
#define	BRECIS_VPM_GET_CALL_CLASS	_IOR('B', 0x3b, int)
#define	BRECIS_VPM_GET_TONE_2100HZ	_IOR('B', 0x3c, int)
#define	BRECIS_VPM_TONE_2100HZ_DETECT	_IOR('B', 0x3d, int)
#define	BRECIS_VPM_JOIN_CONFERENCE	_IOWR('B', 0x3e, MSP_VPM_CONFERENCE *)
#define	BRECIS_VPM_RELEASE_TDM_PORT	_IOR('B', 0x3f, int)
#define	BRECIS_VPM_QUIT_CONFERENCE	_IOWR('B', 0x40, MSP_VPM_CONFERENCE *)

/*
 * Some Caller ID related defines.
 * Not available for manipulation by any ioctls
 * (yet), but that day will have to come.
 */
#define	MSP_VPM_CID_US			0	/* USA */
#define	MSP_VPM_CID_JA			1	/* Japan */

typedef enum
{
	BRECIS_VPM_NEAR_END_2100_TONE_START = 1,
	BRECIS_VPM_NEAR_END_2100_TONE_PHASE_REVERSE_START,
	BRECIS_VPM_NEAR_END_2100_TONE_STOP,
	BRECIS_VPM_FAR_END_2100_TONE_START,
	BRECIS_VPM_FAR_END_2100_TONE_PHASE_REVERSE_START,
	BRECIS_VPM_FAR_END_2100_TONE_STOP
} BRECIS_VPM_TONE_2100HZ_DET_EVENT;

typedef enum
{
    BRECIS_VPM_FAX_DETECTED = 1,
    BRECIS_VPM_DATA_DETECTED
} BRECIS_VPM_CALL_CLASSIFIER_DET_EVENT;

typedef enum
{
    BRECIS_VPM_2100HZ_TONE = 12
} BRECIS_VPM_TONE_2100HZ_ID;

#endif /* BRECIS_MSP_VE_H */
